from typing import TYPE_CHECKING, Any

from bgpy.simulation_engine.ann_containers import RecvQueue

if TYPE_CHECKING:
    from bgpy.shared.enums import Relationships
    from bgpy.simulation_engine.announcement import Announcement as Ann
    from bgpy.simulation_framework import Scenario

    from .bgp import BGP


def seed_ann(self: "BGP", ann: "Ann") -> None:
    """Seeds an announcement at this AS

    Useful hook function used in BGPSec
    and later hopefully in the API for ROAs
    """

    # Ensure we aren't replacing anything
    err = f"Seeding conflict {ann} {self.local_rib.get(ann.prefix)}"
    assert self.local_rib.get(ann.prefix) is None, err
    # Seed by placing in the local rib
    self.local_rib.add_ann(ann)


def receive_ann(self: "BGP", ann: "Ann", accept_withdrawals: bool = False) -> None:
    """Function for recieving announcements, adds to recv_q"""

    if getattr(ann, "withdraw", False) and not accept_withdrawals:
        raise NotImplementedError(f"Policy can't handle withdrawals {self.name}")
    self.recv_q.add_ann(ann)


def process_incoming_anns(
    self: "BGP",
    *,
    from_rel: "Relationships",
    propagation_round: int,
    scenario: "Scenario",
    reset_q: bool = True,
) -> None:
    """Process all announcements that were incoming from a specific rel"""

    # For each prefix, get all anns recieved
    for prefix, ann_list in self.recv_q.items():
        # Get announcement currently in local rib
        current_ann: Ann | None = self.local_rib.get(prefix)
        og_ann = current_ann

        # Seeded Ann will never be overriden, so continue
        if getattr(current_ann, "seed_asn", None) is not None:
            continue
        # For each announcement that was incoming
        for new_ann in ann_list:
            # Make sure there are no loops
            # In ROV subclass also check roa validity
            if self._valid_ann(new_ann, from_rel):
                new_ann_processed = self._copy_and_process(new_ann, from_rel)

                current_ann = self._get_best_ann_by_gao_rexford(
                    current_ann, new_ann_processed
                )

        # This is a new best ann. Process it and add it to the local rib
        if og_ann != current_ann:
            assert current_ann, "mypy type check"
            assert current_ann.seed_asn in (None, self.as_.asn), "Seed ASN is wrong"
            # Save to local rib
            self.local_rib.add_ann(current_ann)

    self._reset_q(reset_q)


def _valid_ann(
    self: "BGP",
    ann: "Ann",
    recv_relationship: "Relationships",
) -> bool:
    """Determine if an announcement is valid or should be dropped"""

    # BGP Loop Prevention Check
    return self.as_.asn not in ann.as_path


def _copy_and_process(
    self: "BGP",
    ann: "Ann",
    recv_relationship: "Relationships",
    overwrite_default_kwargs: dict[Any, Any] | None = None,
) -> "Ann":
    """Deep copies ann and modifies attrs

    Prepends AS to AS Path and sets recv_relationship
    """

    kwargs: dict[str, Any] = {
        "as_path": (self.as_.asn, *ann.as_path),
        "recv_relationship": recv_relationship,
        "seed_asn": None,
    }

    if overwrite_default_kwargs:
        kwargs.update(overwrite_default_kwargs)
    # Don't use a dict comp here for speed
    return ann.copy(overwrite_default_kwargs=kwargs)


def _reset_q(self: "BGP", reset_q: bool) -> None:
    """Resets the recieve q"""

    if reset_q:
        self.recv_q = RecvQueue()



from frozendict import frozendict

from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import BGP
from bgpy.simulation_framework import ScenarioConfig, SubprefixHijack
from bgpy.tests.engine_tests.utils import EngineTestConfig

from .as_graph_info_000 import as_graph_info_000

desc = (
    "Subprefix hijack with BGP Simple"
    "Valley Free (Gao Rexford) Demonstration\n"
    "import policy\n"
    "AS 9, prefix, shows customer > peer\n"
    "AS 9, subprefix, shows peer > provider\n"
    "AS 11, prefix, shows shortest AS path\n"
    "AS 5 and AS 8, subprefix, tiebreaker by lowest ASN\n"
    "export policy\n"
    "AS 10, subprefix, shows anns from providers only export to customers\n"
    "AS 9, subprefix, shows anns from peers only export to customers\n"
    "(All ASes show exporting to cusotmers)\n"
    "hidden hijack\n"
    "AS 12 shows a hidden hijack\n"
)

ex_config_002 = EngineTestConfig(
    name="ex_002_subprefix_hijack_bgp_simple_gao_rexford_demo",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=SubprefixHijack,
        BasePolicyCls=BGP,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict(),
    ),
    as_graph_info=as_graph_info_000,
)



from .aspa import ASPA, ASPAFull, ASRA, ASRAFull
from .bgp import BGP, BGPFull
from .bgpisec import (
    BGPiSecTransitive,
    BGPiSecTransitiveOnlyToCustomers,
    BGPiSecTransitiveProConID,
    BGPiSec,
    ProviderConeID,
    BGPiSecTransitiveFull,
    BGPiSecTransitiveOnlyToCustomersFull,
    BGPiSecTransitiveProConIDFull,
    BGPiSecFull,
    ProviderConeIDFull,
)
from .bgpsec import BGPSec, BGPSecFull
from .custom_attackers import (
    FirstASNStrippingPrefixASPAAttacker,
    ShortestPathPrefixASPAAttacker,
)
from .edge_filter import EdgeFilter, EdgeFilterFull, ROVEdgeFilter, ROVEdgeFilterFull
from .enforce_first_as import (
    EnforceFirstAS,
    EnforceFirstASFull,
    ROVEnforceFirstAS,
    ROVEnforceFirstASFull,
)
from .only_to_customers import OnlyToCustomers, OnlyToCustomersFull
from .path_end import PathEnd, PathEndFull
from .policy import Policy
from .rov import ROV, PeerROV, PeerROVFull, ROVFull
from .rovpp import (
    ROVPPV1Lite,
    ROVPPV1LiteFull,
    ROVPPV2ImprovedLite,
    ROVPPV2ImprovedLiteFull,
    ROVPPV2Lite,
    ROVPPV2LiteFull,
)

__all__ = [
    "BGP",
    "BGPFull",
    "Policy",
    "PeerROV",
    "PeerROVFull",
    "ROV",
    "ROVFull",
    "BGPiSecTransitive",
    "BGPiSecTransitiveOnlyToCustomers",
    "BGPiSecTransitiveProConID",
    "BGPiSec",
    "ProviderConeID",
    "BGPiSecTransitiveFull",
    "BGPiSecTransitiveOnlyToCustomersFull",
    "BGPiSecTransitiveProConIDFull",
    "BGPiSecFull",
    "ProviderConeIDFull",
    "BGPSecFull",
    "BGPSec",
    "EdgeFilter",
    "EdgeFilterFull",
    "ROVEdgeFilter",
    "ROVEdgeFilterFull",
    "EnforceFirstAS",
    "EnforceFirstASFull",
    "ROVEnforceFirstAS",
    "ROVEnforceFirstASFull",
    "OnlyToCustomers",
    "OnlyToCustomersFull",
    "PathEnd",
    "PathEndFull",
    "ROVPPV1Lite",
    "ROVPPV1LiteFull",
    "ROVPPV2Lite",
    "ROVPPV2LiteFull",
    "ROVPPV2ImprovedLite",
    "ROVPPV2ImprovedLiteFull",
    "ASPA",
    "ASPAFull",
    "ASRA",
    "ASRAFull",
    "ShortestPathPrefixASPAAttacker",
    "FirstASNStrippingPrefixASPAAttacker",
]



from typing import TYPE_CHECKING

from bgpy.simulation_engine.policies.rov import ROVFull

from .rovpp_v1_lite import ROVPPV1Lite

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann


class ROVPPV1LiteFull(ROVPPV1Lite, ROVFull):
    """An Policy that deploys ROV++V1 Lite as defined in the ROV++ paper, and
    has withdrawals, ribs in and out

    ROV++ Improved Deployable Defense against BGP Hijacking
    """

    name: str = "ROV++V1 Lite Full"

    def _add_blackholes_tolocal_rib(self, blackholes: tuple["Ann", ...]) -> None:
        """Adds all blackholes to the local RIB"""

        for blackhole in blackholes:
            existing_ann = self.local_rib.get(blackhole.prefix)
            # Don't overwrite valid existing announcements
            if existing_ann is None:
                self.local_rib.add_ann(blackhole)
            elif existing_ann.invalid_by_roa:
                # If you need this feature, please email jfuruness@gmail.com
                # Although I think this case would be very rare that no one will
                # ever come across it so I won't bother with implementing it right now
                raise NotImplementedError(
                    "Need to handle withdrawals for this"
                    " if you need this feature, please email jfuruness@gmail.com"
                )
                self.local_rib.add_ann(blackhole)  # type: ignore



from ipaddress import ip_network
from typing import TYPE_CHECKING, Optional

from roa_checker import ROA

from bgpy.shared.enums import Prefixes, Relationships, Timestamps
from bgpy.simulation_framework.scenarios.scenario import Scenario

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann
    from bgpy.simulation_engine import BaseSimulationEngine


class VictimsPrefix(Scenario):
    """A Scenario with only the victims prefix

    This is meant to be used as a parent class, so it still has
    attackers, even though only the victims prefix is used
    """

    def _get_announcements(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple["Ann", ...]:
        """Returns a valid prefix announcement

        for subclasses of this EngineInput, you can set AnnCls equal to
        something other than Announcement
        """

        anns = list()
        for victim_asn in self.victim_asns:
            anns.append(
                self.scenario_config.AnnCls(
                    prefix=Prefixes.PREFIX.value,
                    next_hop_asn=victim_asn,
                    as_path=(victim_asn,),
                    timestamp=Timestamps.VICTIM.value,
                    seed_asn=victim_asn,
                    roa_valid_length=True,
                    roa_origin=victim_asn,
                    recv_relationship=Relationships.ORIGIN,
                )
            )
        return tuple(anns)

    def _get_roas(
        self,
        *,
        announcements: tuple["Ann", ...] = (),
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple[ROA, ...]:
        """Returns a tuple of ROAs"""

        return tuple(
            [ROA(ip_network(Prefixes.PREFIX.value), x) for x in self.victim_asns]
        )



from typing import TYPE_CHECKING

from .edge_filter import EdgeFilter

if TYPE_CHECKING:
    from bgpy.shared.enums import Relationships
    from bgpy.simulation_engine import Announcement as Ann


class ROVEdgeFilter(EdgeFilter):
    """Prevents edge ASes from paths containing ASNs they don't own (w/ROV)"""

    name: str = "ROV + EdgeFilter"

    def _valid_ann(self, ann: "Ann", from_rel: "Relationships") -> bool:
        """ROV+EdgeFilter"""

        return False if ann.invalid_by_roa else super()._valid_ann(ann, from_rel)



import csv
import pickle
from pathlib import Path

from bgpy.simulation_engine import BaseSimulationEngine
from bgpy.simulation_framework import GraphDataAggregator, Scenario
from bgpy.utils import EngineRunner


class EngineTester(EngineRunner):
    """Tests an engine run"""

    def __init__(
        self,
        *args,
        overwrite: bool = False,
        compare_graph_data: bool = False,
        **kwargs,
    ) -> None:
        """Regarding the compare_graph_data kwarg:

        There was quite a debate on whether or not the engine tester
        should compare metrics. There are a lot of cons for comparing
        metrics here. Every single time the metrics change, all of the
        test data needs to be overwritten. This means even if you add
        a new metric, the test data needs to be overwritten, even if it
        is unrelated to the metrics, and you'll need to double check the
        ground truth of everything once again. metrics are also very
        finnicky, if JSON or CSV layouts change, they are prone to breakage
        without actually being a true failure. If the layout changes then
        those files change in git, even though no code has been modified.
        On top of that, different
        simulation_frameworks for different projects have different
        metrick trackers, some of which are quite different, some of which
        have no metrics at all. It also breaks
        the single responsibility of these being engine tests, to test the
        engine, not the simulation_framework. metric tracker could also be
        tested with unit tests.

        As for the cons, some tests would require a full round of propagation
        for a true metric tracker test. While the majority of the feedback
        was to remove it, one project was already using it. In light of that,
        being able to compare metrics in this test class will remain a feature,
        but by default it will be off. I've left an example to show how
        other projects could implement their own versions of it.
        """

        self.overwrite: bool = overwrite
        self.compare_graph_data: bool = compare_graph_data
        super().__init__(*args, **kwargs)

        # Don't store metrics if we don't use them
        if not self.compare_graph_data:

            def noop(*args, **kwargs):
                pass

            self._store_metrics = noop

    def test_engine(self):
        """Tests an engine run

        Takes in a scenario (initialized with adopt ASN, atk and vic ASN,
        and a graph
        The scenario + graph are used to build and seed the engine
        After the engine is seeded, the engine is run
        Data is collected from the engine
        The engine and traceback are output to YAML
        We then compare the current run's traceback and engine
            to the ground truth
        """

        engine, outcomes_yaml, graph_data_aggregator, scenario = self.run_engine()
        # Store engine and traceback YAML
        self._store_gt_data(engine, outcomes_yaml, graph_data_aggregator)
        # Create diagrams before the test can fail
        self._generate_gt_diagrams(scenario, graph_data_aggregator)
        # Compare the YAML's together
        self._compare_data()

    def _store_gt_data(
        self,
        engine: BaseSimulationEngine,
        outcomes: dict[int, int],
        graph_data_aggregator: GraphDataAggregator,
    ) -> None:
        """Stores GROUND TRUTH YAML for the engine, outcomes, and CSV for metrics.

        If ground truth doesn't exist, create it
        """

        # Save engine as ground truth if ground truth doesn't exist
        if not self.engine_ground_truth_path.exists() or self.overwrite:
            self.codec.dump(engine, path=self.engine_ground_truth_path)
        # Save outcomes as ground truth if ground truth doesn't exist
        if not self.outcomes_ground_truth_path.exists() or self.overwrite:
            self.codec.dump(outcomes, path=self.outcomes_ground_truth_path)

        self._store_gt_metrics(graph_data_aggregator)

    def _store_gt_metrics(self, graph_data_aggregator: GraphDataAggregator) -> None:
        """Stores metric ground truth

        For some reason, even with sorting, these files store
        differently, and this the git diffs include these files
        even though we don't even track them anymore

        To combat this, we now compare metrics, and only overwrite
        if the metrics are actually different

        This function is unfortunately stupid complicated to accomplish this
        """

        # Save metrics as ground truth if ground truth doesn't exist
        if (
            not self.graph_data_ground_truth_path_pickle.exists()
            or not self.graph_data_ground_truth_path_csv.exists()
        ):
            graph_data_aggregator.write_data(
                csv_path=self.graph_data_ground_truth_path_csv,
                pickle_path=self.graph_data_ground_truth_path_pickle,
            )
        elif self.overwrite and self.graph_data_guess_path_pickle.exists():
            try:
                self._compare_graph_data_to_gt()
            # Metrics are actually different, write them out
            except AssertionError:
                graph_data_aggregator.write_data(
                    csv_path=self.graph_data_ground_truth_path_csv,
                    pickle_path=self.graph_data_ground_truth_path_pickle,
                )
        elif self.overwrite:
            try:
                # Write to guess first, so that we can compare
                graph_data_aggregator.write_data(
                    csv_path=self.graph_data_guess_path_csv,
                    pickle_path=self.graph_data_guess_path_pickle,
                )
                self._compare_graph_data_to_gt()
            # Metrics are actually different, write them out
            except AssertionError:
                graph_data_aggregator.write_data(
                    csv_path=self.graph_data_ground_truth_path_csv,
                    pickle_path=self.graph_data_ground_truth_path_pickle,
                )

    def _generate_gt_diagrams(
        self, scenario: Scenario, graph_data_aggregator: GraphDataAggregator
    ) -> None:
        """Generates diagrams for ground truth"""

        # Load engines
        engine_gt = self.codec.load(self.engine_ground_truth_path)
        # Load outcomes
        outcomes_gt = self.codec.load(self.outcomes_ground_truth_path)

        static_order = bool(self.conf.as_graph_info.diagram_ranks)
        diagram_obj_ranks = self._get_diagram_obj_ranks(engine_gt)

        # Write ground truth graph
        self.conf.DiagramCls().generate_as_graph(
            engine_gt,
            scenario,
            outcomes_gt,
            f"({self.conf.name} Ground Truth)\n" f"{self.conf.desc}",
            graph_data_aggregator,
            diagram_obj_ranks,
            static_order=static_order,
            path=self.storage_dir / "ground_truth.gv",
            view=False,
            dpi=self.dpi,
        )

    def _compare_data(self) -> None:
        """Compares YAML for ground truth vs guess for engine and outcomes"""

        # Compare Engine
        engine_guess = self.codec.load(self.engine_guess_path)
        engine_gt = self.codec.load(self.engine_ground_truth_path)
        assert engine_guess == engine_gt, f"{self.conf.name} failed engine check"
        # Compare outcomes
        outcomes_guess = self.codec.load(self.outcomes_guess_path)
        outcomes_gt = self.codec.load(self.outcomes_ground_truth_path)
        assert outcomes_guess == outcomes_gt, f"{self.conf.name} failed outcomes check"

        if self.compare_graph_data:
            self._compare_graph_data_to_gt()

    def _compare_graph_data_to_gt(self) -> None:
        # Compare metrics CSV
        with self.graph_data_guess_path_csv.open() as guess_f:
            with self.graph_data_ground_truth_path_csv.open() as ground_truth_f:
                guess_lines = {tuple(x) for x in csv.reader(guess_f)}
                gt_lines = {tuple(x) for x in csv.reader(ground_truth_f)}
                assert gt_lines == guess_lines, self.graph_data_guess_path_csv

        # Compare metrics YAML
        with self.graph_data_guess_path_pickle.open("rb") as f:
            graph_data_guess = pickle.load(f)  # noqa: S301
        with self.graph_data_ground_truth_path_pickle.open("rb") as f:
            graph_data_gt = pickle.load(f)  # noqa: S301

        for graph_category, data_point_dict in graph_data_guess.items():
            for data_point_key, agg_trial_data in data_point_dict.items():
                assert graph_data_gt[graph_category][data_point_key] == agg_trial_data
        for graph_category, data_point_dict in graph_data_gt.items():
            for data_point_key, agg_trial_data in data_point_dict.items():
                assert graph_data_guess[graph_category][data_point_key] == (
                    agg_trial_data
                )

    #########
    # Paths #
    #########

    @property
    def engine_ground_truth_path(self) -> Path:
        """Returns the path to the engine's ground truth YAML"""

        return self.storage_dir / "engine_gt.yaml"

    @property
    def outcomes_ground_truth_path(self) -> Path:
        """Returns the path to the outcomes ground truth YAML"""

        return self.storage_dir / "outcomes_gt.yaml"

    @property
    def graph_data_ground_truth_path_csv(self) -> Path:
        """Returns the path to the metrics ground truth YAML"""

        return self.storage_dir / "graph_data_gt.csv"

    @property
    def graph_data_ground_truth_path_pickle(self) -> Path:
        """Returns the path to the metrics ground truth YAML"""

        return self.storage_dir / "graph_data_gt.pickle"



from .enforce_first_as import EnforceFirstAS
from .enforce_first_as_full import EnforceFirstASFull
from .rov_enforce_first_as import ROVEnforceFirstAS
from .rov_enforce_first_as_full import ROVEnforceFirstASFull


__all__ = [
    "EnforceFirstASFull",
    "EnforceFirstAS",
    "ROVEnforceFirstASFull",
    "ROVEnforceFirstAS",
]



import abc
from abc import ABC, abstractmethod
from functools import cached_property as cached_property
from ipaddress import IPv4Network, IPv6Network
from typing import Any

from _typeshed import Incomplete
from roa_checker import ROA as ROA
from roa_checker import ROAChecker as ROAChecker

from bgpy.as_graphs import AS as AS
from bgpy.shared.enums import SpecialPercentAdoptions as SpecialPercentAdoptions
from bgpy.simulation_engine import (
    Announcement as Ann,
)
from bgpy.simulation_engine import (
    BaseSimulationEngine as BaseSimulationEngine,
)
from bgpy.simulation_engine import (
    Policy as Policy,
)
from bgpy.simulation_framework.scenarios.scenario_config import (
    ScenarioConfig as ScenarioConfig,
)

class Scenario(ABC, metaclass=abc.ABCMeta):
    min_propagation_rounds: int
    scenario_config: Incomplete
    percent_adoption: Incomplete
    attacker_asns: Incomplete
    victim_asns: Incomplete
    adopting_asns: Incomplete
    announcements: Incomplete
    roas: Incomplete
    ordered_prefix_subprefix_dict: Incomplete
    def __init__(
        self,
        *,
        scenario_config: ScenarioConfig,
        percent_adoption: float | SpecialPercentAdoptions = 0,
        engine: BaseSimulationEngine | None = None,
        attacker_asns: frozenset[int] | None = None,
        victim_asns: frozenset[int] | None = None,
        adopting_asns: frozenset[int] | None = None,
    ) -> None: ...
    def _get_attacker_asns(
        self,
        override_attacker_asns: frozenset[int] | None,
        prev_attacker_asns: frozenset[int] | None,
        engine: BaseSimulationEngine | None,
    ) -> frozenset[int]: ...
    def _get_possible_attacker_asns(
        self,
        engine: BaseSimulationEngine,
        percent_adoption: float | SpecialPercentAdoptions,
    ) -> frozenset[int]: ...
    def _get_victim_asns(
        self,
        override_victim_asns: frozenset[int] | None,
        prev_victim_asns: frozenset[int] | None,
        engine: BaseSimulationEngine | None,
    ) -> frozenset[int]: ...
    def _get_possible_victim_asns(
        self,
        engine: BaseSimulationEngine,
        percent_adoption: float | SpecialPercentAdoptions,
    ) -> frozenset[int]: ...
    def _get_adopting_asns(
        self,
        override_adopting_asns: frozenset[int] | None,
        adopting_asns: frozenset[int] | None,
        engine: BaseSimulationEngine | None,
    ) -> frozenset[int]: ...
    def _get_randomized_adopting_asns(
        self, engine: BaseSimulationEngine
    ) -> frozenset[int]: ...
    @cached_property
    def _default_adopters(self) -> frozenset[int]: ...
    @cached_property
    def _default_non_adopters(self) -> frozenset[int]: ...
    @cached_property
    def _preset_asns(self) -> frozenset[int]: ...
    @property
    def untracked_asns(self) -> frozenset[int]: ...
    @property
    def _untracked_asns(self) -> frozenset[int]: ...
    def setup_engine(self, engine: BaseSimulationEngine) -> None: ...
    def get_policy_cls(self, as_obj: AS) -> type[Policy]: ...
    @abstractmethod
    def _get_announcements(
        self, *, engine: BaseSimulationEngine | None = None
    ) -> tuple[Ann, ...]: ...
    def _get_roas(
        self,
        *,
        announcements: tuple[Ann, ...] = (),
        engine: BaseSimulationEngine | None = None,
    ) -> tuple[ROA, ...]: ...
    def pre_aggregation_hook(
        self,
        engine: BaseSimulationEngine,
        percent_adopt: float | SpecialPercentAdoptions,
        trial: int,
        propagation_round: int,
    ) -> None: ...
    def post_propagation_hook(
        self,
        engine: BaseSimulationEngine,
        percent_adopt: float | SpecialPercentAdoptions,
        trial: int,
        propagation_round: int,
    ) -> None: ...
    def _add_roa_info_to_anns(
        self,
        *,
        announcements: tuple[Ann, ...] = (),
        engine: BaseSimulationEngine | None = None,
    ) -> tuple[Ann, ...]: ...
    def _get_roa_checker(self) -> ROAChecker: ...
    def _get_roa_origin(
        self, roa_checker: ROAChecker, prefix: IPv4Network | IPv6Network, origin: int
    ) -> int | None: ...
    def _get_roa_valid_length(
        self, roa_checker: ROAChecker, prefix: IPv4Network | IPv6Network, origin: int
    ) -> bool | None: ...
    def _get_ordered_prefix_subprefix_dict(self): ...
    def __to_yaml_dict__(self) -> dict[Any, Any]: ...
    @classmethod
    def __from_yaml_dict__(cls, dct, yaml_tag): ...



from typing import TYPE_CHECKING

from bgpy.shared.enums import Relationships
from bgpy.simulation_engine.policies.rovpp.v1.rovpp_v1_lite import ROVPPV1Lite

if TYPE_CHECKING:
    from bgpy.as_graphs import AS
    from bgpy.simulation_engine import Announcement as Ann


class ROVPPV2Lite(ROVPPV1Lite):
    """An Policy that deploys ROV++V2 Lite as defined in the ROV++ paper

    ROV++ Improved Deployable Defense against BGP Hijacking

    """

    name: str = "ROV++V2 Lite"

    def _policy_propagate(
        self,
        neighbor: "AS",
        ann: "Ann",
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> bool:
        """Deals with blackhole propagation

        If ann is a blackhole, it must be recv from peers/providers and must
        be sent only to customers
        """

        # If blackhole and subprefix or non-routed
        if ann.rovpp_blackhole:
            if self._send_competing_hijack_allowed(ann, propagate_to):
                self._process_outgoing_ann(neighbor, ann, propagate_to, send_rels)
            return True
        else:
            return False

    def _send_competing_hijack_allowed(
        self, ann: "Ann", propagate_to: Relationships
    ) -> bool:
        """You can send blackhole to customers if from peer/provider
        and either subprefix or non routed
        """
        return (
            # From peer/provider
            ann.recv_relationship
            in [Relationships.PEERS, Relationships.PROVIDERS, Relationships.ORIGIN]
            # Sending to customers
            and propagate_to == Relationships.CUSTOMERS
            # subprefix or non routed (don't send blackholes for prefixes)
            and (not ann.roa_valid_length or not ann.roa_routed)
        )



from bgpy.simulation_engine.policies.rov.rov_full import ROVFull

from .enforce_first_as import EnforceFirstAS


class EnforceFirstASFull(EnforceFirstAS, ROVFull):
    """An Policy that deploys EnforceFirstAS and has withdrawals, ribs in and out"""

    name: str = "Enforce-First-AS Full"



import bz2
import shutil
from datetime import datetime, timedelta
from functools import cached_property
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import cast

import requests
from bs4 import BeautifulSoup as Soup

from bgpy.as_graphs.base import ASGraphCollector
from bgpy.shared.constants import bgpy_logger
from bgpy.shared.exceptions import NoCAIDAURLError


class CAIDAASGraphCollector(ASGraphCollector):
    """Downloads relationships from CAIDA and caches file"""

    def _run(self) -> Path:
        """Downloads relationships into a file

        https://publicdata.caida.org/datasets/as-relationships/serial-2/

        Can specify a download time if you want to download an older dataset
        if cache is True it uses the downloaded file that was cached
        """

        if not self.cache_path.exists():
            bgpy_logger.info("No caida graph cached. Caching...")
            # Create a temporary dir to write to
            with TemporaryDirectory() as tmp_dir:
                # Path to bz2 download
                bz2_path: Path = Path(tmp_dir) / "download.bz2"
                # Download Bz2
                self._download_bz2_file(self._get_url(self.dl_time), bz2_path)
                self._unzip_and_write_to_cache(bz2_path)
        return self.cache_path

    @cached_property
    def default_dl_time(self) -> datetime:
        """Returns default DL time.

        For most things, we download from 4 days ago
        And for collectors, time must be divisible by 4/8
        """

        # 10 days because sometimes caida takes a while to upload
        # 7 days ago was actually not enough
        dl_time: datetime = datetime.now() - timedelta(days=10)
        return dl_time.replace(hour=0, minute=0, second=0, microsecond=0)

    #################
    # Request funcs #
    #################

    def _get_url(self, dl_time: datetime) -> str:
        """Gets urls to download relationship files"""

        # Api url
        prepend: str = "http://data.caida.org/datasets/as-relationships/serial-2/"
        # Gets all URLs. Keeps only the link for the proper download time
        urls = [
            prepend + x
            for x in self._get_hrefs(prepend)
            if dl_time.strftime("%Y%m01") in x
        ]
        if len(urls) > 0:
            return str(urls[0])
        else:  # pragma: no cover
            raise NoCAIDAURLError("No Urls")

    def _get_hrefs(self, url: str) -> list[str]:
        """Returns hrefs from a tags at a given url"""

        try:
            # Query URL
            with requests.get(url, stream=True, timeout=30) as r:
                # Check for errors
                r.raise_for_status()
                # Get soup
                soup = Soup(r.text, "html.parser")
                # Extract hrefs from a tags
                rv = [
                    x.get("href") for x in soup.select("a") if x.get("href") is not None
                ]
                return cast(list[str], rv)
        except requests.exceptions.ReadTimeout as e:
            bgpy_logger.exception(f"Failed to get {url} due to {e}")
            raise

    #########################
    # File formatting funcs #
    #########################

    def _download_bz2_file(self, url: str, bz2_path: Path) -> None:
        """Downloads bz2 file from caida"""

        # https://stackoverflow.com/a/39217788/8903959
        # Download the file
        with requests.get(url, stream=True, timeout=5) as r:
            r.raise_for_status()
            with bz2_path.open("wb") as f:
                shutil.copyfileobj(r.raw, f)

    def _unzip_and_write_to_cache(self, bz2_path: Path) -> None:
        """Unzips bz2 file and writes to cache"""

        # Unzip and read
        with bz2.open(bz2_path, mode="rb") as bz2_f, self.cache_path.open("w") as txt_f:
            for line in bz2_f:
                # Must decode the bytes into strings and strip
                txt_f.write(line.decode().strip() + "\n")



from .scenario_config import ScenarioConfig  # isort: skip
from .scenario import Scenario  # isort: skip

from .custom_scenarios import (
    AccidentalRouteLeak,
    FirstASNStrippingPrefixHijack,
    ForgedOriginPrefixHijack,
    NonRoutedPrefixHijack,
    NonRoutedSuperprefixHijack,
    NonRoutedSuperprefixPrefixHijack,
    PrefixHijack,
    ShortestPathPrefixHijack,
    SubprefixHijack,
    SuperprefixPrefixHijack,
    ValidPrefix,
    VictimsPrefix,
)

__all__ = [
    "Scenario",
    "ScenarioConfig",
    "AccidentalRouteLeak",
    "PrefixHijack",
    "SubprefixHijack",
    "NonRoutedPrefixHijack",
    "NonRoutedSuperprefixHijack",
    "NonRoutedSuperprefixPrefixHijack",
    "ForgedOriginPrefixHijack",
    "FirstASNStrippingPrefixHijack",
    "ShortestPathPrefixHijack",
    "SuperprefixPrefixHijack",
    "ValidPrefix",
    "VictimsPrefix",
]



from dataclasses import dataclass

from bgpy.as_graphs import ASGraph, ASGraphInfo, CAIDAASGraph
from bgpy.simulation_engine import BaseSimulationEngine, SimulationEngine
from bgpy.simulation_framework.as_graph_analyzers import (
    ASGraphAnalyzer,
    BaseASGraphAnalyzer,
)
from bgpy.simulation_framework.graph_data_aggregator.graph_data_aggregator import (
    GraphDataAggregator,
)
from bgpy.simulation_framework.scenarios import ScenarioConfig

from .diagram import Diagram


@dataclass(frozen=True, slots=True)
class EngineRunConfig:
    """Configuration info for a single engine run

    Useful for tests, API calls, or just running a single configuration
    """

    name: str
    desc: str
    scenario_config: ScenarioConfig
    as_graph_info: ASGraphInfo
    requires_provider_cones: bool = False
    ASGraphCls: type[ASGraph] = CAIDAASGraph
    # Only concrete classes are allowed, but not going to bother with a protocol here
    SimulationEngineCls: type[BaseSimulationEngine] = SimulationEngine  # type: ignore
    GraphDataAggregatorCls: type[GraphDataAggregator] = GraphDataAggregator
    ASGraphAnalyzerCls: type[BaseASGraphAnalyzer] = ASGraphAnalyzer
    DiagramCls: type[Diagram] = Diagram



from .accidental_route_leak import AccidentalRouteLeak
from .non_routed import (
    NonRoutedPrefixHijack,
    NonRoutedSuperprefixHijack,
    NonRoutedSuperprefixPrefixHijack,
)
from .post_rov import (
    FirstASNStrippingPrefixHijack,
    ForgedOriginPrefixHijack,
    ShortestPathPrefixHijack,
    SuperprefixPrefixHijack,
)
from .pre_rov import PrefixHijack, SubprefixHijack
from .valid_prefix import ValidPrefix
from .victims_prefix import VictimsPrefix

__all__ = [
    "AccidentalRouteLeak",
    "PrefixHijack",
    "SubprefixHijack",
    "NonRoutedPrefixHijack",
    "NonRoutedSuperprefixHijack",
    "NonRoutedSuperprefixPrefixHijack",
    "ForgedOriginPrefixHijack",
    "FirstASNStrippingPrefixHijack",
    "ShortestPathPrefixHijack",
    "SuperprefixPrefixHijack",
    "ValidPrefix",
    "VictimsPrefix",
]



from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

from yamlable import YamlAble, yaml_info, yaml_info_decorate

# https://stackoverflow.com/a/57005931/8903959
if TYPE_CHECKING:
    from bgpy.as_graphs import ASGraph
    from bgpy.simulation_framework import Scenario


@yaml_info(yaml_tag="BaseSimulationEngine")
class BaseSimulationEngine(YamlAble, ABC):
    """AS Graph wrapper that supports announcement propogation

    This class must be first setup with the _setup function
    This resets all the ASes to their base state, and changes
    the classes to be adopting specific defensive policies
    Then the run function can be called, and propagation occurs
    """

    def __init_subclass__(cls, *args, **kwargs):
        """This method essentially creates a list of all subclasses
        This is allows us to easily assign yaml tags
        """

        super().__init_subclass__(*args, **kwargs)
        # Fix this later once the system test framework is updated
        yaml_info_decorate(cls, yaml_tag=cls.__name__)

    def __init__(
        self,
        as_graph: "ASGraph",
        # Useful for C++ Engine
        cached_as_graph_tsv_path: Path | None = None,
        ready_to_run_round: int = -1,
    ) -> None:
        """Saves read_to_run_rund attr and inits superclass"""

        self.as_graph = as_graph
        # Useful for C++ version
        self.cached_as_graph_tsv_path: Path | None = cached_as_graph_tsv_path
        # This indicates whether or not the simulator has been set up for a run
        # We use a number instead of a bool so that we can indicate for
        # each round whether it is ready to run or not
        self.ready_to_run_round: int = ready_to_run_round

    def __eq__(self, other) -> bool:
        """Returns if two simulators contain the same BGPDAG's"""

        if isinstance(other, BaseSimulationEngine):
            rv = self.as_graph.as_dict == other.as_graph.as_dict
            assert isinstance(rv, bool), "Make mypy happy"
            return rv
        else:
            return NotImplemented

    ###############
    # Setup funcs #
    ###############

    @abstractmethod
    def setup(self, scenario: "Scenario") -> None:
        """Sets AS classes and seeds announcements"""

        raise NotImplementedError

    #####################
    # Propagation funcs #
    #####################

    @abstractmethod
    def run(
        self, propagation_round: int = 0, scenario: Optional["Scenario"] = None
    ) -> None:
        """Propogates announcements and ensures proper setup"""

        raise NotImplementedError

    ##############
    # Yaml funcs #
    ##############

    @abstractmethod
    def __to_yaml_dict__(self) -> dict[str, Any]:
        """This optional method is called when you call yaml.dump()"""

        raise NotImplementedError

    @classmethod
    @abstractmethod
    def __from_yaml_dict__(
        cls: type["BaseSimulationEngine"], dct: dict[str, Any], yaml_tag: Any
    ) -> "BaseSimulationEngine":
        """This optional method is called when you call yaml.load()"""

        raise NotImplementedError



from frozendict import frozendict

from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import BGP
from bgpy.simulation_framework import ScenarioConfig, ValidPrefix
from bgpy.tests.engine_tests.utils import EngineTestConfig

from .as_graph_info_000 import as_graph_info_000

desc = "Valid prefix with BGP Simple"

ex_config_000 = EngineTestConfig(
    name="ex_000_valid_prefix_bgp_simple",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=ValidPrefix,
        BasePolicyCls=BGP,
        override_attacker_asns=frozenset(),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict(),
    ),
    as_graph_info=as_graph_info_000,
)



from frozendict import frozendict

from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import BGP, PathEnd
from bgpy.simulation_framework import ScenarioConfig, ShortestPathPrefixHijack
from bgpy.tests.engine_tests.utils import EngineTestConfig

from .as_graph_info_000 import as_graph_info_000

desc = (
    "shortest path export all attack against pathend simple\n"
    "PathEnd checks the end of the path for valid providers\n"
    "so anything beyond the third AS is not protected"
)

ex_config_017 = EngineTestConfig(
    name="ex_017_shortest_path_export_all_pathend_simple",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=ShortestPathPrefixHijack,
        BasePolicyCls=BGP,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict(
            {
                1: PathEnd,
                ASNs.VICTIM.value: PathEnd,
            }
        ),
    ),
    as_graph_info=as_graph_info_000,
)



from typing import TYPE_CHECKING

from bgpy.shared.enums import Relationships
from bgpy.simulation_engine.policies.rov import ROV

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann


class ASPA(ROV):
    """A Policy that deploys ASPA and ASPA Records

    We adopt from ROV since deploying ASPA makes no sense without ROV

    We experimented with adding a cache to the provider_check
    but this has a negligible impact on performance

    Removing the path reversals sped up performance by about 5%
    but made the code a lot less readable and deviated from the RFC,
    so we decided to forgo those as well

    This is now up to date with the V18 proposal
    """

    name: str = "ASPA"

    def _valid_ann(self, ann: "Ann", from_rel: Relationships) -> bool:
        """Returns False if from peer/customer when aspa is set"""

        assert len(set(ann.as_path)) == len(ann.as_path), "We deal with prepending"

        # Note: This first if check has to be removed if you want to implement
        # route server to RS-Client behaviour
        if not self._next_hop_valid(ann):
            return False
        # Most ASes recieve anns from providers (moved here for speed)
        elif from_rel.value == Relationships.PROVIDERS.value:
            return self._downstream_check(ann, from_rel)
        elif from_rel.value in (
            Relationships.CUSTOMERS.value,
            Relationships.PEERS.value,
        ):
            return self._upstream_check(ann, from_rel)
        else:
            raise NotImplementedError("Should never reach here")

    def _next_hop_valid(self, ann: "Ann") -> bool:
        """Ensures the next hop is the first ASN in the AS-Path

        This behavior would need to be modified for route servers
        """

        return ann.next_hop_asn == ann.as_path[0]

    def _upstream_check(self, ann: "Ann", from_rel: "Relationships") -> bool:
        """ASPA upstream check"""

        # Upstream check
        if len(ann.as_path) == 1:
            return super()._valid_ann(ann, from_rel)
        # For every adopting ASPA AS in the path,
        # The next ASN in the path must be in their providers list
        # Since this is checking from customers

        # 4. If max_up_ramp < N, the procedure halts with the outcome "Invalid".
        elif self._get_max_up_ramp_length(ann) < len(ann.as_path):
            return False

        # ASPA valid or unknown
        return super()._valid_ann(ann, from_rel)

    def _get_max_up_ramp_length(self, ann: "Ann") -> int:
        """See desc

        Determine the maximum up-ramp length as I, where I is the minimum
        index for which authorized(A(I), A(I+1)) returns "Not Provider+".  If
        there is no such I, the maximum up-ramp length is set equal to the
        AS_PATH length N.  This parameter is defined as max_up_ramp

        The up-ramp starts at AS(1) and each hop AS(i) to AS(i+1) represents
        Customer and Provider peering relationship. [i.e they reverse the path]
        """

        reversed_path = ann.as_path[::-1]

        for i in range(len(reversed_path) - 1):
            if not self._provider_check(reversed_path[i], reversed_path[i + 1]):
                return i + 1
        return len(ann.as_path)

    def _downstream_check(self, ann: "Ann", from_rel: "Relationships") -> bool:
        """ASPA downstream check"""

        # 4. If max_up_ramp + max_down_ramp < N,
        # the procedure halts with the outcome "Invalid".
        max_up_ramp = self._get_max_up_ramp_length(ann)
        max_down_ramp = self._get_max_down_ramp_length(ann)
        if max_up_ramp + max_down_ramp < len(ann.as_path):
            return False

        # ASPA Valid or Unknown (but not invalid)
        return super()._valid_ann(ann, from_rel)

    def _get_max_down_ramp_length(self, ann: "Ann") -> int:
        """See desc

        Similarly, the maximum down-ramp length can be determined as N - J +
        1 where J is the maximum index for which authorized(A(J), A(J-1))
        returns "Not Provider+".  If there is no such J, the maximum down-
        ramp length is set equal to the AS_PATH length N.  This parameter is
        defined as max_down_ramp.

        In the down-ramp, each pair AS(j) to
        AS(j-1) represents Customer and Provider peering relationship
        """

        reversed_path = ann.as_path[::-1]

        # We want the max J, so start at the end of the reversed Path
        # This is the most efficient way to traverse this
        for i in range(len(reversed_path) - 1, 0, -1):
            if not self._provider_check(reversed_path[i], reversed_path[i - 1]):
                # Must add one due to zero indexing in python, vs 1 indexing in RFC
                J = i + 1
                return len(reversed_path) - J + 1
        return len(ann.as_path)

    def _provider_check(self, asn1: int, asn2: int) -> bool:
        """Returns False if asn2 is not in asn1's provider_asns, AND asn1 adopts ASPA

        This also essentially can take the place of the "hop check" listed in
        ASPA RFC section 5 in ASPA v16
        or in ASPAv18 it takes the place of the provider auth func

        False indicates Not Provider+
        True indicates No Attestation or Provider+
        """

        cur_as_obj = self.as_.as_graph.as_dict[asn1]
        if isinstance(cur_as_obj.policy, ASPA):
            next_as_obj = self.as_.as_graph.as_dict[asn2]
            if next_as_obj.asn not in cur_as_obj.provider_asns:
                return False
        return True



import re
from pathlib import Path
from typing import Iterable

from requests_cache import CachedSession

from bgpy.shared.constants import SINGLE_DAY_CACHE_DIR
from bgpy.shared.enums import ASGroups, InAdoptingASNs, Outcomes, Plane
from bgpy.simulation_framework.graph_data_aggregator.graph_category import GraphCategory


def get_all_graph_categories() -> Iterable[GraphCategory]:
    """Returns all possible metric key combos"""

    for plane in [Plane.DATA]:
        for as_group in [ASGroups.ALL_WOUT_IXPS]:
            for outcome in [x for x in Outcomes if x != Outcomes.UNDETERMINED]:
                for in_adopting_asns_enum in list(InAdoptingASNs):
                    yield GraphCategory(
                        plane=plane,
                        as_group=as_group,
                        outcome=outcome,
                        in_adopting_asns=in_adopting_asns_enum,
                    )


def get_country_asns(
    country_code: str,
    requests_cache_path: Path = SINGLE_DAY_CACHE_DIR / "requests_cache.db",
) -> list[int]:
    """
    Returns all the ASNs associated with a specific region. A region is given by a
    two-letter country code defined in the ISO 3166-1 standard. More details about
    country codes can be found here: https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2.
    """

    with CachedSession(str(requests_cache_path)) as session:
        response = session.get(
            "https://stat.ripe.net/data/country-asns/"
            f"data.json?resource={country_code.lower()}&lod=1"
        )
        assert response.status_code == 200

        # Extract both routed and non-routed ASNs using regex
        country_info = response.json()["data"]["countries"][0]
        asns = re.findall(r"AsnSingle\((\d+)\)", country_info["routed"])
        asns += re.findall(r"AsnSingle\((\d+)\)", country_info["non_routed"])

        return [int(asn) for asn in asns]



from typing import TYPE_CHECKING

from .aspa import ASPA

if TYPE_CHECKING:
    from bgpy.enums import Relationships
    from bgpy.simulation_engine.announcement import Announcement as Ann


class ASRA(ASPA):
    """ASRA: Esentially ASPA and checking neighbors at every AS together

    Originally we had this policy coded for ASPAwN, an extension of ASPA that
    we had developed independently of ASRA in parallel. Upon publishing an ASPA
    evaluation paper, we received feedback from an ASPA RFC author and converted
    ASPAwN into ASRA

    That being said, ASRA is not yet finalized, and at the time of this writing the
    draft wasn't even published. This policy assumes ASRA3 records (not specifying
    the neighbor type) and also assumes the strict algorithm.

    I also don't know why the draft is made out to be so complicated. I'm leaving
    the algorithm we used for ASPAwN below, as it's vastly simpler than the ASRA
    algorithm listed in the RFC but has the same effect. I've contacted an ASRA
    author and asked them to simplify it. Once ASRA becomes finalized I'll change
    the code below.
    """

    name = "ASRA"

    def _valid_ann(self, ann: "Ann", from_rel: "Relationships") -> bool:  # type: ignore
        """Combines ASPA valid and checking neighbors at every AS"""

        as_path = ann.as_path
        as_dict = self.as_.as_graph.as_dict
        for i, asn in enumerate(as_path):
            # Get the AS object for the current AS in the AS Path
            asra_as_obj = as_dict[asn]
            # If the AS is an ASRA AS
            if isinstance(asra_as_obj.policy, ASRA):
                # Check that both of it's neighbors are in the valid next hops
                for neighbor_index in (i - 1, i + 1):
                    # Can't use try except IndexError here, since -1 is a valid index
                    if 0 <= neighbor_index < len(as_path):
                        neighbor_asn = as_path[neighbor_index]
                        if neighbor_asn not in asra_as_obj.neighbor_asns:
                            return False
        rv = super()._valid_ann(ann, from_rel)
        assert isinstance(rv, bool), "mypy"
        return rv



from bgpy.simulation_engine.policies.rov.rov_full import ROVFull

from .bgpisec_transitive_pro_con_id import BGPiSecTransitiveProConID


class BGPiSecTransitiveProConIDFull(BGPiSecTransitiveProConID, ROVFull):
    """BGP-iSec with withdrawals, ribs in and out"""

    name = "BGP-iSec Transitive + ProConID Full"



from frozendict import frozendict

from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import BGP, OnlyToCustomers
from bgpy.simulation_framework import AccidentalRouteLeak, ScenarioConfig
from bgpy.tests.engine_tests.utils import EngineTestConfig

from .as_graph_info_000 import as_graph_info_000

desc = (
    "accidental route leak against OnlyToCustomersSimple\n"
    "This policy sets the only_to_customers attribute"
    "specified in RFC 9234 \n"
    "which protects against simple route leaks"
)

ex_config_019 = EngineTestConfig(
    name="ex_019_route_leak_otc_simple",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=AccidentalRouteLeak,
        BasePolicyCls=BGP,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict(
            {
                1: OnlyToCustomers,
                2: OnlyToCustomers,
            }
        ),
    ),
    as_graph_info=as_graph_info_000,
)



from typing import TYPE_CHECKING, Any, Optional

from bgpy.shared.exceptions import GaoRexfordError
from bgpy.simulation_engine.policies.rov import ROV

if TYPE_CHECKING:
    from bgpy.as_graphs import AS
    from bgpy.shared.enums import Relationships
    from bgpy.simulation_engine.announcement import Announcement as Ann


class BGPSec(ROV):
    """Represents BGPSec

    Since there are no real world implementations,
    we assume a secure path preference of security third,
    which is in line with the majority of users
    for the survey results in "A Survey of Interdomain Routing Policies"
    https://www.cs.bu.edu/~goldbe/papers/survey.pdf

    Also - this adopts from ROV since it's extremely unlikely that an AS
    would deploy BGPSec without first deploying ROV
    """

    name = "BGPsec"

    def seed_ann(self, ann: "Ann") -> None:
        """Seeds announcement at this AS and initializes BGPSec path"""

        # This ann is valid, add the bgpsec as path
        if ann.as_path == (self.as_.asn,):
            ann = ann.copy({"bgpsec_as_path": ann.as_path})
        super().seed_ann(ann)

    def _policy_propagate(
        self,
        neighbor: "AS",
        ann: "Ann",
        propagate_to: "Relationships",
        send_rels: set["Relationships"],
    ) -> bool:
        """Sets BGPSec fields when propagating

        If sending to bgpsec, set next_as and keep bgpsec_as_path
        otherwise clear out both fields
        """

        if isinstance(neighbor.policy, BGPSec):
            next_asn = neighbor.asn
            path = ann.bgpsec_as_path
        else:
            next_asn = None
            path = ()
        send_ann = ann.copy({"bgpsec_next_asn": next_asn, "bgpsec_as_path": path})
        self._process_outgoing_ann(neighbor, send_ann, propagate_to, send_rels)
        return True

    # Mypy doesn't understand the superclass
    def _copy_and_process(
        self,
        ann: "Ann",
        recv_relationship: "Relationships",
        overwrite_default_kwargs: dict[Any, Any] | None = None,
    ) -> "Ann":
        """Sets the bgpsec_as_path.

        prepends ASN if valid, otherwise clears
        """
        if ann.bgpsec_valid(self.as_.asn):
            bgpsec_as_path = (self.as_.asn, *ann.bgpsec_as_path)
        else:
            bgpsec_as_path = ()

        if overwrite_default_kwargs is None:
            overwrite_default_kwargs = {}

        overwrite_default_kwargs["bgpsec_as_path"] = overwrite_default_kwargs.get(
            "bgpsec_as_path", bgpsec_as_path
        )

        return super()._copy_and_process(
            ann, recv_relationship, overwrite_default_kwargs
        )

    def _get_best_ann_by_bgpsec(
        self, current_ann: "Ann", new_ann: "Ann"
    ) -> Optional["Ann"]:
        current_valid = current_ann.bgpsec_valid(self.as_.asn)
        new_valid = new_ann.bgpsec_valid(self.as_.asn)

        if current_valid and not new_valid:
            return current_ann
        elif not current_valid and new_valid:
            return new_ann
        else:
            return None

    def _get_best_ann_by_gao_rexford(
        self,
        current_ann: Optional["Ann"],
        new_ann: "Ann",
    ) -> "Ann":
        """Determines if the new ann > current ann by Gao Rexford"""

        assert new_ann is not None, "New announcement can't be None"

        if current_ann is None:
            return new_ann
        else:
            # Inspiration for this func refactor came from bgpsecsim
            # for func in self._gao_rexford_funcs:
            #     best_ann = func(current_ann, new_ann)
            #     if best_ann is not None:
            #         assert isinstance(best_ann, Ann), "mypy type check"
            #         return best_ann

            # Having this dynamic like above is literally 7x slower, resulting
            # in bottlenecks. Gotta do it the ugly way unfortunately
            ann = self._get_best_ann_by_local_pref(current_ann, new_ann)
            if ann:
                return ann
            else:
                ann = self._get_best_ann_by_as_path(current_ann, new_ann)
                if ann:
                    return ann
                else:
                    ann = self._get_best_ann_by_bgpsec(current_ann, new_ann)
                    if ann:
                        return ann
                    else:
                        return self._get_best_ann_by_lowest_neighbor_asn_tiebreaker(
                            current_ann, new_ann
                        )
            raise GaoRexfordError("No ann was chosen")



from .as_graph_analyzer import ASGraphAnalyzer
from .base_as_graph_analyzer import BaseASGraphAnalyzer
from .interception_as_graph_analyzer import InterceptionASGraphAnalyzer

__all__ = ["BaseASGraphAnalyzer", "ASGraphAnalyzer", "InterceptionASGraphAnalyzer"]



import warnings
from collections import deque
from typing import TYPE_CHECKING, Optional

from bgpy.shared.constants import bgpy_logger
from bgpy.shared.enums import (
    Prefixes,
    Relationships,
    SpecialPercentAdoptions,
    Timestamps,
)
from bgpy.simulation_engine import (
    ASPA,
    ASRA,
    BGP,
    ROV,
    ASPAFull,
    ASRAFull,
    BGPFull,
    BGPiSec,
    BGPiSecFull,
    BGPiSecTransitive,
    BGPiSecTransitiveFull,
    BGPiSecTransitiveOnlyToCustomers,
    BGPiSecTransitiveOnlyToCustomersFull,
    BGPiSecTransitiveProConID,
    BGPiSecTransitiveProConIDFull,
    BGPSec,
    BGPSecFull,
    EdgeFilter,
    EdgeFilterFull,
    EnforceFirstAS,
    EnforceFirstASFull,
    OnlyToCustomers,
    OnlyToCustomersFull,
    PathEnd,
    PathEndFull,
    PeerROV,
    PeerROVFull,
    Policy,
    ProviderConeID,
    ProviderConeIDFull,
    ROVEdgeFilter,
    ROVEdgeFilterFull,
    ROVEnforceFirstAS,
    ROVEnforceFirstASFull,
    ROVFull,
    ROVPPV1Lite,
    ROVPPV1LiteFull,
    ROVPPV2ImprovedLite,
    ROVPPV2ImprovedLiteFull,
    ROVPPV2Lite,
    ROVPPV2LiteFull,
    ShortestPathPrefixASPAAttacker,
)
from bgpy.simulation_framework.scenarios.custom_scenarios.pre_rov import PrefixHijack
from bgpy.simulation_framework.scenarios.custom_scenarios.victims_prefix import (
    VictimsPrefix,
)

from .forged_origin_prefix_hijack import ForgedOriginPrefixHijack

if TYPE_CHECKING:
    from bgpy.as_graphs import AS
    from bgpy.simulation_engine import Announcement as Ann
    from bgpy.simulation_engine import BaseSimulationEngine


class ShortestPathPrefixHijack(VictimsPrefix):
    """Shortest path allowed by defense set by AdoptPolicyCls against a prefix"""

    RequiredASPAAttackerCls = ShortestPathPrefixASPAAttacker

    def _get_announcements(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple["Ann", ...]:
        """Returns the two announcements seeded for this engine input

        This engine input is for a prefix hijack,
        consisting of a valid prefix and invalid prefix with path manipulation
        """

        # First get the victims prefix
        victim_anns = super()._get_announcements(engine=engine)
        attacker_anns = self._get_shortest_path_attacker_anns(engine=engine)
        return victim_anns + attacker_anns

    def _get_shortest_path_attacker_anns(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple["Ann", ...]:
        """Returns announcements for the shortest path attacker"""

        if self.scenario_config.AdoptPolicyCls in self.pre_rov_policy_classes:
            # Types are a bit messed due to mixins
            return self._get_prefix_attacker_anns(engine=engine)  # type: ignore
        elif self.scenario_config.AdoptPolicyCls in self.rov_policy_classes:
            # Types are a bit messed due to mixins
            return self._get_forged_origin_attacker_anns(engine=engine)  # type: ignore
        elif self.scenario_config.AdoptPolicyCls in self.pathend_policy_classes:
            return self._get_pathend_attack_anns(engine=engine)
        elif self.scenario_config.AdoptPolicyCls in self.aspa_policy_classes:
            return self._get_aspa_attack_anns(engine=engine)
        elif self.scenario_config.AdoptPolicyCls in self.asra_policy_classes:
            return self._get_aspa_attack_anns(
                engine=engine, require_aspa_attacker_cls=False
            )
        elif self.scenario_config.AdoptPolicyCls in self.bgpisec_policy_classes:
            # See post_propagation_hook - this is where the attack takes place
            # This must happen this way since bgpisec requires an actual ann
            # whereas ASPA just requires a "potential/plausible" ann
            return ()
        else:
            raise NotImplementedError(
                "Need to code shortest path attack against "
                f"{self.scenario_config.AdoptPolicyCls}"
            )

    # Using mixins instead of weird subclassing, SLF001 is wrong here (private vars)
    _get_prefix_attacker_anns = PrefixHijack._get_prefix_attacker_anns  # noqa: SLF001
    _get_forged_origin_attacker_anns = (
        # Using mixins instead of weird subclassing, SLF001 is wrong here (private vars)
        ForgedOriginPrefixHijack._get_forged_origin_attacker_anns  # noqa: SLF001
    )

    def _get_pathend_attack_anns(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple["Ann", ...]:
        """Get shortest path undetected by Path-End"""

        assert engine, "mypy"
        if len(self.victim_asns) > 1:
            raise NotImplementedError

        root_as_obj = engine.as_graph.as_dict[next(iter(self.victim_asns))]
        root_asn = root_as_obj.asn
        shortest_valid_path: tuple[int, ...] | None = None
        for first_provider in root_as_obj.providers:
            # You only need legit origin and their provider, you don't need three
            # for secondary_provider in first_provider.providers:
            #     return (secondary_provider.asn, first_provider.asn, root_asn)
            shortest_valid_path = (first_provider.asn, root_asn)
            break
        if not shortest_valid_path:
            # Some ASes don't have providers, and are stubs that are peered
            for first_peer in root_as_obj.peers:
                shortest_valid_path = (first_peer.asn, root_asn)
                break

        if not shortest_valid_path:
            # Strange case but it could happen
            for first_customer in root_as_obj.customers:
                shortest_valid_path = (first_customer.asn, root_asn)
                break

        if shortest_valid_path is None:
            warnings.warn(
                "Shortest path against pathend is none? "
                "This should only happen at full adoption...",
                stacklevel=2,
            )
            shortest_valid_path = ()

        assert shortest_valid_path is not None, "mypy"
        anns = list()
        for attacker_asn in self.attacker_asns:
            anns.append(
                self.scenario_config.AnnCls(
                    prefix=Prefixes.PREFIX.value,
                    as_path=(attacker_asn, *shortest_valid_path),
                    timestamp=Timestamps.ATTACKER.value,
                    next_hop_asn=attacker_asn,
                    seed_asn=attacker_asn,
                )
            )
        return tuple(anns)

    def _get_aspa_attack_anns(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
        require_aspa_attacker_cls: bool = True,
    ) -> tuple["Ann", ...]:
        """Get shortest path undetected by ASPA"""

        if len(self.victim_asns) > 1:
            raise NotImplementedError

        if require_aspa_attacker_cls:
            self._validate_required_aspa_attacker_cls()

        assert engine, "mypy"

        shortest_valid_path = self._find_shortest_valley_free_non_adopting_path(
            root_asn=next(iter(self.victim_asns)), engine=engine
        )
        anns = list()
        for attacker_asn in self.attacker_asns:
            # There are cases where the attacker is a part of this
            # We add the attacker later so just remove it here
            current_shortest_valid_path = tuple(
                [x for x in shortest_valid_path if x != attacker_asn]
            )
            anns.append(
                self.scenario_config.AnnCls(
                    prefix=Prefixes.PREFIX.value,
                    as_path=(attacker_asn, *current_shortest_valid_path),
                    timestamp=Timestamps.ATTACKER.value,
                    next_hop_asn=attacker_asn,
                    seed_asn=attacker_asn,
                )
            )
        return tuple(anns)

    def _validate_required_aspa_attacker_cls(self) -> None:
        """Validates that AttackerBasePolicyCls is set correctly against ASPA"""

        if self.scenario_config.AttackerBasePolicyCls != self.RequiredASPAAttackerCls:
            raise ValueError(
                "For a shortest path export all attack against ASPA, "
                "scenario_config.AttackerAdoptPolicyCls must be set to "
                "ASPAShortestPathPrefixAttacker, which you can import like "
                "from bgpy.simulation_engine import ShortestPathPrefixASPAAttacker"
            )

    def _find_shortest_valley_free_non_adopting_path(
        self,
        root_asn: int,
        engine: "BaseSimulationEngine",
    ) -> tuple[int, ...]:
        """Finds the shortest non adopting path from the root asn

        Announcements from customers > peers > providers, since
        policies like ASPA and bgp-isec would reject announcements
        that are already going to customers, etc. So even if the path
        is longer, it's better to be accepted by going to a provider
        """

        AdoptPolicyCls = self.scenario_config.AdoptPolicyCls

        root_as = engine.as_graph.as_dict[root_asn]

        # {ASN: as_path to get here}
        # NOTE: I used to have AS as the key, but weakref.Proxy isn't hashable
        # https://stackoverflow.com/a/68273386/8903959
        visited = dict()

        # First, use BFS on provider relationships
        queue: deque[tuple[AS, tuple[int, ...]]] = deque([(root_as, (root_as.asn,))])
        while queue:
            as_, as_path = queue.popleft()
            if as_.asn not in visited:
                if not issubclass(self.get_policy_cls(as_), AdoptPolicyCls):
                    return as_path
                visited[as_.asn] = as_path
                for provider_as in engine.as_graph.as_dict[as_.asn].providers:
                    if provider_as.asn not in visited:
                        queue.append((provider_as, (provider_as.asn, *as_path)))

        # Then, go in order of provider relationships
        # This is a nice optimization, since the dictionary maintains order
        # and BFS increments in order of distance
        for visited_asn, as_path in visited.copy().items():
            visited_as = engine.as_graph.as_dict[visited_asn]
            for peer_as in visited_as.peers:
                if not issubclass(self.get_policy_cls(peer_as), AdoptPolicyCls):
                    return (peer_as.asn, *as_path)
                elif peer_as.asn not in visited:
                    visited[peer_as.asn] = (peer_as.asn, *as_path)

        # At this point, if we still haven't found it, it's a customer
        # relationship (or doesn't exist).
        # From here, the proper way to do things would be to use some modified
        # djikstras algorithm
        # But it's important to note that this is the uncommon case
        # since this would only occur if all of the input clique is adopting.
        # If that were the case, djikstras algorithm would have a very bad runtime
        # since 99% of ASes would be adopting
        # Additionally, this function only runs once per trial
        # for all these reasons, I'm not going to implement a modified djikstras,
        # which may be prone to error, and just do it the naive way, which is much
        # less error prone
        # To do this, I'll  simply iterate through all remaining ASes, and then sort
        # them and return the shortest AS path (or None)

        # BFS code removed here

        # The above commented out implementation basically did BFS using
        # customer relationships from every provider within the provider + peer
        # cone. The problem with that is you end up rechecking the same ASN
        # many different times, and you essentially are doing BFS 1000 times
        # due to the size of the provider + peer cone.
        # The only way to avoid rechecking the ASes, and avoid blowing up the RAM
        # is actually to use propagation ranks, since the graph is essentially ordered
        # As to why it seems to occur so frequently, when the input clique typically
        # should have non adopting ASNs, it's because about .2% of the CAIDA graph
        # gets disconnected. That means 99.8% of the time, a victim can reach the
        # input clique. But .998 ^ 1000 trials means ~13% chance that the victim
        # can reach the input clique every time. That's why we had to fix this,
        # even though in theory the problem should rarely occur.

        # So this is in the end doing something very similar to propagation
        # however, there are a few differences that I think make this worth keeping
        # 1. It's wayyyy faster than copying announcements and function call overhead
        #    especially with non-simple policies
        # 2. We can use this to modify any hijacks (whereas if this used propagation,
        #    it would be it's own scenario class
        # 3. The algo stops when it reaches a non adopting ASN, and the overwhelming
        #    majority of the time, that's going to be in the provider cone of the root
        #    asn, which is usually < 1000 ASes, which is going to be very fast
        # anyways, on with the propagation lol

        msg = "propagation ranks don't need reversal"
        assert engine.as_graph.propagation_ranks[-1][0].propagation_rank != 0, msg
        non_adopting_customer_asns = set()
        for propagation_rank in reversed(engine.as_graph.propagation_ranks):
            for as_obj in propagation_rank:
                shortest_provider_path: tuple[int, ...] | None = None
                for provider_asn in as_obj.provider_asns:
                    provider_path = visited.get(provider_asn)
                    if provider_path is not None and (
                        shortest_provider_path is None
                        or len(provider_path) < len(shortest_provider_path)
                    ):
                        shortest_provider_path = provider_path
                # relevant to root ASN
                if shortest_provider_path:
                    new_as_path = (as_obj.asn, *shortest_provider_path)
                    old_as_path = visited.get(as_obj.asn)
                    if not old_as_path or len(old_as_path) > len(new_as_path):
                        visited[as_obj.asn] = new_as_path
                    if not issubclass(self.get_policy_cls(as_obj), AdoptPolicyCls):
                        non_adopting_customer_asns.add(as_obj.asn)

        # Sort through non adopting customer ASNs to find shortest path
        if non_adopting_customer_asns:
            non_adopting_customer_paths = {
                asn: visited[asn] for asn in non_adopting_customer_asns
            }
            sorted_non_adopting_customer_paths = sorted(
                non_adopting_customer_paths.items(), key=lambda x: len(x[1])
            )
            best_asn, best_as_path = sorted_non_adopting_customer_paths[0]
            return best_as_path
        else:
            warnings.warn(
                "Shortest path against ASPA is none? "
                "This should only happen at full adoption...",
                stacklevel=2,
            )
            return ()

    def post_propagation_hook(
        self,
        engine: "BaseSimulationEngine",
        percent_adopt: float | SpecialPercentAdoptions,
        trial: int,
        propagation_round: int,
    ) -> None:
        """Performs Shortest path prefix hijack against BGPiSec classes

        For this attack, since bgp-isec transitive attributes are present,
        an actual announcement in a local RIB of an AS must be chosen
        (this assumes the attacker can see any local RIB through looking glass
        servers or route collectors, etc). This is unlike the theoretical shortest
        path like against ASPA (or in other words - when attacking ASPA you
        want the shortest plausible path. When attacking bgpisec you need the
        shortest path that exists).

        To do this, we iterate through all ASes and choose an announcement that
        has the shortest path in a local RIB where the last AS on the path is not
        adopting. Then the attacker appends their ASN and we repropagate.
        We could do this using algorithms and such depending on the class, but
        that can be a future improvement. For now I'm going for accuracy.

        A caveat for selecting the best shortest path - the attacker should always
        prefer announcements which have no OTC attributes over announcements with
        OTC attributes, since the attacker would then be detected less when
        propagating to providers.

        Additionally, much like an AccidentalRouteLeak, previously the attacker would
        simply modify their local RIB and then propagate again. However - this
        has some drawbacks. Then the attacker must deploy BGPFull
        (that uses withdrawals) and the entire graph has to propagate again.
        BGPFull (and subclasses of it) are MUCH slower than BGP due to all the extra
        computations for withdrawals, RIBsIn, RIBsOut, etc. Additionally,
        propagating a second round after the ASGraph is __already__ full
        is wayyy more expensive than propagating when the AS graph is empty.

        Instead, we now get the announcement that the attacker needs to leak
        after the first round of propagating the valid prefix.
        Then we clear the graph, seed those announcements, and propagate again
        This way, we avoid needing BGPFull (since the graph has been cleared,
        there is no need for withdrawals), and we avoid propagating a second
        time after the graph is alrady full.

        Since this simulator treats each propagation round as if it all happens
        at once, this is possible.
        """

        if self.scenario_config.AdoptPolicyCls not in self.bgpisec_policy_classes:
            return
        elif propagation_round == 0:
            # Force their to be two rounds of propagation
            # Can't set this in the class var since you only want it to apply here
            if self.scenario_config.propagation_rounds < 2:
                raise ValueError("Please set ScenarioConfig.propagation_rounds to 2")

            announcements: list[Ann] = list(self.announcements)

            # Find the best ann for attacker to fake with
            best_ann: Ann | None = None
            for _asn, as_obj in engine.as_graph.as_dict.items():
                # Search for an AS that doesn't next_hop signature
                if not isinstance(as_obj.policy, BGPiSecTransitive):
                    for ann in as_obj.policy.local_rib.values():
                        if best_ann is None:
                            best_ann = ann
                        # Prefer anns without OTC, always
                        elif ann.only_to_customers and not best_ann.only_to_customers:
                            continue
                        # Lastly prefer shorter paths
                        elif len(ann.as_path) < len(best_ann.as_path):
                            best_ann = ann
            if not best_ann:
                # NOTE: may be possible due to the 1% being all disconnected ASes
                # or when valid ann is in one of those disconnected ASes
                # this happens if you run 1k trials
                bgpy_logger.info(f"Couldn't find best_ann at {percent_adopt}% adoption")
                # When this occurs, use victim's ann to at least do forged-origin
                victim_as_obj = engine.as_graph.as_dict[next(iter(self.victim_asns))]
                for ann in victim_as_obj.policy.local_rib.values():
                    best_ann = ann

            assert best_ann, "mypy"
            # Add fake announcements
            assert self.attacker_asns, "You must select at least 1 AS to leak"
            for attacker_asn in self.attacker_asns:
                announcements.append(
                    best_ann.copy(
                        {
                            "as_path": (attacker_asn, *best_ann.as_path),
                            "recv_relationship": Relationships.ORIGIN,
                            "seed_asn": attacker_asn,
                            "timestamp": Timestamps.ATTACKER.value,
                            "next_hop_asn": attacker_asn,
                        }
                    )
                )

            # Reset the engine for the next run
            self.announcements = tuple(announcements)
            self.setup_engine(engine)
            engine.ready_to_run_round = 1
        elif propagation_round > 1:
            raise NotImplementedError

    ######################
    # Policy class lists #
    ######################

    @property
    def pre_rov_policy_classes(self) -> frozenset[type[Policy]]:
        """These are policy classes that are susceptible to pre_rov attacks

        such as prefix hijack, subprefix hijack
        """

        return frozenset(
            {
                BGP,
                BGPFull,
                OnlyToCustomers,
                OnlyToCustomersFull,
                EdgeFilter,
                EdgeFilterFull,
                EnforceFirstAS,
                EnforceFirstASFull,
            }
        )

    @property
    def rov_policy_classes(self) -> frozenset[type[Policy]]:
        """These are policy classes that are susceptible to forged-origin attacks"""

        return frozenset(
            {
                BGPSec,
                BGPSecFull,
                ROVEdgeFilter,
                ROVEdgeFilterFull,
                ROVEnforceFirstAS,
                ROVEnforceFirstASFull,
                ROV,
                ROVFull,
                PeerROV,
                PeerROVFull,
                ROVPPV1Lite,
                ROVPPV1LiteFull,
                ROVPPV2Lite,
                ROVPPV2LiteFull,
                ROVPPV2ImprovedLite,
                ROVPPV2ImprovedLiteFull,
                ProviderConeID,
                ProviderConeIDFull,
            }
        )

    @property
    def pathend_policy_classes(self) -> frozenset[type[Policy]]:
        """Policies susceptible to a Pathend Shortest Path attack

        i.e the third hop is where the attacker is placed
        """

        return frozenset({PathEnd, PathEndFull})

    @property
    def aspa_policy_classes(self) -> frozenset[type[Policy]]:
        """Policies susceptible to an ASPA Shortest Path attack

        i.e. shortest contiguous chain of ASPA providers ending in non adopter
        with forged-origin hijack sent to customers
        """

        return frozenset({ASPA, ASPAFull})

    @property
    def asra_policy_classes(self) -> frozenset[type[Policy]]:
        """Policies susceptible to an ASRA Shortest Path attack

        i.e. shortest contiguous chain of ASPA providers ending in non adopter
        """

        return frozenset({ASRA, ASRAFull})

    @property
    def bgpisec_policy_classes(self) -> frozenset[type[Policy]]:
        """Policies susceptible to an BGP-iSec Shortest Path attack

        i.e. shortest contiguous chain of BGP-iSec ASes ending in non adopter
        that uses actual announcements rather than plausible ones
        """

        return frozenset(
            {
                BGPiSecTransitive,
                BGPiSecTransitiveOnlyToCustomers,
                BGPiSecTransitiveProConID,
                BGPiSec,
                BGPiSecTransitiveFull,
                BGPiSecTransitiveOnlyToCustomersFull,
                BGPiSecTransitiveProConIDFull,
                BGPiSecFull,
            }
        )



import warnings
from typing import TYPE_CHECKING, Optional

from bgpy.as_graphs.base.as_graph.customer_cone_funcs import _get_cone_size_helper
from bgpy.shared.constants import bgpy_logger
from bgpy.shared.enums import (
    ASGroups,
    Relationships,
    SpecialPercentAdoptions,
    Timestamps,
)
from bgpy.simulation_framework.scenarios.scenario_config import ScenarioConfig

from .victims_prefix import VictimsPrefix

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann
    from bgpy.simulation_engine import BaseSimulationEngine


class AccidentalRouteLeak(VictimsPrefix):
    """An accidental route leak of a valid prefix"""

    min_propagation_rounds: int = 2

    def __init__(
        self,
        *,
        scenario_config: ScenarioConfig,
        percent_adoption: float | SpecialPercentAdoptions = 0,
        engine: Optional["BaseSimulationEngine"] = None,
        attacker_asns: frozenset[int] | None = None,
        victim_asns: frozenset[int] | None = None,
        adopting_asns: frozenset[int] | None = None,
    ):
        assert engine, "Need engine for customer cones"
        self._attackers_customer_cones_asns: set[int] = set()
        super().__init__(
            scenario_config=scenario_config,
            percent_adoption=percent_adoption,
            engine=engine,
            attacker_asns=attacker_asns,
            victim_asns=victim_asns,
            adopting_asns=adopting_asns,
        )
        self.validate_attacker_subcategory()

    def validate_attacker_subcategory(self) -> None:
        """Validates that the attacker's subcategory/ASGroup can leak"""

        if (
            self.scenario_config.attacker_subcategory_attr in self.warning_as_groups
            and not self.scenario_config.override_attacker_asns
        ):
            msg = (
                "You used the ASGroup of "
                f"{self.scenario_config.attacker_subcategory_attr} "
                f"for your scenario {self.__class__.__name__}, "
                f"but {self.__class__.__name__} can't leak from stubs. "
                "To suppress this warning, override warning_as_groups. "
                "To change the ASGroup to something other than stubs, you can "
                " set attacker_subcategory_attr=ASGroups.MULTIHOMED.value, "
                " in the scenario config after importing like "
                "from bgpy.shared.enums import ASGroups"
            )
            warnings.warn(msg, RuntimeWarning, stacklevel=2)

    # Just returns customer cone
    _get_cone_size_helper = _get_cone_size_helper

    def post_propagation_hook(
        self,
        engine: "BaseSimulationEngine",
        percent_adopt: float | SpecialPercentAdoptions,
        trial: int,
        propagation_round: int,
    ) -> None:
        """Causes an accidental route leak

        Changes the valid prefix to be received from a customer
        so that in the second propagation round, the AS will export to all
        relationships

        NOTE: the old way of doing this was to simply alter the attackers
        local RIB and then propagate again. However - this has some drawbacks
        Then the attacker must deploy BGPFull (that uses withdrawals) and
        the entire graph has to propagate again. BGPFull (and subclasses
        of it) are MUCH slower than BGP due to all the extra
        computations for withdrawals, RIBsIn, RIBsOut, etc. Additionally,
        propagating a second round after the ASGraph is __already__ full
        is wayyy more expensive than propagating when the AS graph is empty.

        Instead, we now get the announcement that the attacker needs to leak
        after the first round of propagating the valid prefix.
        Then we clear the graph, seed those announcements, and propagate again
        This way, we avoid needing BGPFull (since the graph has been cleared,
        there is no need for withdrawals), and we avoid propagating a second
        time after the graph is alrady full.

        Since this simulator treats each propagation round as if it all happens
        at once, this is possible.

        Additionally, you could also do the optimization in the first propagation
        round to only propagate from ASes that can reach the attacker. But we'll
        forgo this for now for simplicity.
        """

        if propagation_round == 0:
            announcements: list[Ann] = list(self.announcements)
            assert self.attacker_asns, "You must select at least 1 AS to leak"
            for attacker_asn in self.attacker_asns:
                if not engine.as_graph.as_dict[attacker_asn].policy.local_rib:
                    bgpy_logger.warning(
                        "Attacker did not recieve announcement, can't leak."
                    )
                for _prefix, ann in engine.as_graph.as_dict[
                    attacker_asn
                ].policy.local_rib.items():
                    announcements.append(
                        ann.copy(
                            {
                                "recv_relationship": Relationships.ORIGIN,
                                "seed_asn": attacker_asn,
                                "timestamp": Timestamps.ATTACKER.value,
                            }
                        )
                    )
            self.announcements = tuple(announcements)
            self.setup_engine(engine)
            engine.ready_to_run_round = 1
        elif propagation_round > 1:
            raise NotImplementedError

    def _get_attacker_asns(
        self,
        override_attacker_asns: frozenset[int] | None,
        attacker_asns: frozenset[int] | None,
        engine: Optional["BaseSimulationEngine"],
    ) -> frozenset[int]:
        """Gets attacker ASNs, overriding the valid prefix which has no attackers

        There is a very rare case where the attacker can not perform the route leak
        due to a disconnection, which happens around .1% in the CAIDA topology.
        In theory - you could just look at the provider cone of the victim,
        and then the peers of that provider cone (and of the victim itself), and
        then the customer cones of all of those ASes to get the list of possible
        valid attackers. However, we consider the attacker being unable to attack
        in extremely rare cases a valid result, and thus do not change the random
        selection. Doing so would also be a lot slower for a very extreme edge case
        """

        assert engine, "Need engine for attacker customer cones"
        attacker_asns = super()._get_attacker_asns(
            override_attacker_asns, attacker_asns, engine
        )
        # Stores customer cones of attacker ASNs
        # used in untrackable func and when selecting victims
        for attacker_asn in attacker_asns:
            self._attackers_customer_cones_asns.update(
                self._get_cone_size_helper(
                    engine.as_graph.as_dict[attacker_asn],
                    dict(),
                ),
            )
        return attacker_asns

    def _get_possible_victim_asns(
        self,
        engine: "BaseSimulationEngine",
        percent_adoption: float | SpecialPercentAdoptions,
    ) -> frozenset[int]:
        """Returns possible victim ASNs, defaulted from config

        Modified to not allow victims to be in attackers customer cone,
        since if a victim is the customer of leaker, it's not really a leak
        """

        possible_asns = super()._get_possible_victim_asns(engine, percent_adoption)
        # Remove attacker's customer conesfrom possible victims
        possible_asns = possible_asns.difference(self._attackers_customer_cones_asns)
        return possible_asns

    @property
    def warning_as_groups(self) -> frozenset[str]:
        """Returns a frozenset of ASGroups that should raise a warning"""

        return frozenset(
            [
                ASGroups.STUBS_OR_MH.value,
                ASGroups.STUBS.value,
                ASGroups.ALL_WOUT_IXPS.value,
            ]
        )

    @property
    def _untracked_asns(self) -> frozenset[int]:
        """Returns ASNs that shouldn't be tracked by the metric tracker

        By default just the default adopters and non adopters
        however for the route leak, we don't want to track the customers of the
        leaker, since you can not "leak" to your own customers
        """

        return super()._untracked_asns | self._attackers_customer_cones_asns



from dataclasses import dataclass
from typing import ClassVar

from bgpy.utils import EngineRunConfig


@dataclass(frozen=True, slots=True)
class EngineTestConfig(EngineRunConfig):
    """Configuration info for the test suite"""

    _used_names: ClassVar[set[str]] = set()

    def __post_init__(self):
        """Names are used as folder names for testing, can't have duplicates"""

        if self.name in EngineTestConfig._used_names:
            raise ValueError(f"The name '{self.name}' is already in use.")
        EngineTestConfig._used_names.add(self.name)



from typing import TYPE_CHECKING, Optional

from bgpy.simulation_engine.policies.custom_attackers.first_asn_stripping_prefix_aspa_attacker import (  # noqa: E501
    FirstASNStrippingPrefixASPAAttacker,
)
from bgpy.simulation_framework.scenarios.custom_scenarios.victims_prefix import (
    VictimsPrefix,
)

from .shortest_path_prefix_hijack import ShortestPathPrefixHijack

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann
    from bgpy.simulation_engine import BaseSimulationEngine


class FirstASNStrippingPrefixHijack(ShortestPathPrefixHijack):
    """An extension of the shortest path hijack that strips the first ASN"""

    RequiredASPAAttackerCls = FirstASNStrippingPrefixASPAAttacker

    def _get_announcements(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple["Ann", ...]:
        """Returns the two announcements seeded for this engine input

        This engine input is for a prefix hijack,
        consisting of a valid prefix and invalid prefix with path manipulation
        """

        # First get the victims prefix
        # Using mixins instead of weird subclassing, SLF001 is wrong here (private vars)
        victim_anns = VictimsPrefix._get_announcements(  # noqa: SLF001
            self, engine=engine
        )
        attacker_anns = self._get_first_asn_stripped_attacker_anns(engine=engine)
        return victim_anns + attacker_anns

    def _get_first_asn_stripped_attacker_anns(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple["Ann", ...]:
        attacker_anns = self._get_shortest_path_attacker_anns(engine=engine)
        stripped_anns: list[Ann] = list()
        for ann in attacker_anns:
            # Remove the attacker's ASN
            if len(ann.as_path) > 1:
                stripped_ann = ann.copy({"as_path": ann.as_path[1:]})
            # Can't remove the first asn in a path length of 1
            else:
                stripped_ann = ann
            stripped_anns.append(stripped_ann)
        return tuple(stripped_anns)



from pathlib import Path

from frozendict import frozendict

from bgpy.as_graphs.base import (
    ASGraph,
    ASGraphCollector,
    ASGraphConstructor,
    ASGraphInfo,
    PeerLink,
)
from bgpy.as_graphs.base import CustomerProviderLink as CPLink

from .caida_as_graph import CAIDAASGraph
from .caida_as_graph_collector import CAIDAASGraphCollector


class CAIDAASGraphConstructor(ASGraphConstructor):
    # Add an optional default to ASGraphCollectorCls and ASGraphCls
    def __init__(
        self,
        ASGraphCollectorCls: type[ASGraphCollector] = CAIDAASGraphCollector,
        ASGraphCls: type[ASGraph] = CAIDAASGraph,
        as_graph_collector_kwargs=frozendict(),
        as_graph_kwargs=frozendict(),
        tsv_path: Path | None = None,
        stubs: bool = True,
    ) -> None:
        super().__init__(
            ASGraphCollectorCls,
            ASGraphCls,
            as_graph_collector_kwargs=as_graph_collector_kwargs,
            as_graph_kwargs=as_graph_kwargs,
            tsv_path=tsv_path,
            stubs=stubs,
        )

    ####################
    # Abstract methods #
    ####################
    def _get_as_graph_info(
        self, dl_path: Path, invalid_asns: frozenset[int] = frozenset()
    ) -> ASGraphInfo:
        """Gets AS Graph info from the downloaded file"""

        input_clique_asns: set[int] = set()
        ixp_asns: set[int] = set()
        # Customer provider links
        cp_links: set[CPLink] = set()
        # Peer links
        peer_links: set[PeerLink] = set()

        with dl_path.open() as f:
            for line in f:
                # Get Caida input clique. See paper on site for what this is
                if line.startswith("# input clique"):
                    self._extract_input_clique_asns(
                        line, input_clique_asns, invalid_asns
                    )
                # Get detected Caida IXPs. See paper on site for what this is
                elif line.startswith("# IXP ASes"):
                    self._extract_ixp_asns(line, ixp_asns, invalid_asns)
                # Not a comment, must be a relationship
                elif not line.startswith("#"):
                    # Extract all customer provider pairs
                    if "-1" in line:
                        self._extract_provider_customers(line, cp_links, invalid_asns)
                    # Extract all peers
                    else:
                        self._extract_peers(line, peer_links, invalid_asns)

        return ASGraphInfo(
            customer_provider_links=frozenset(cp_links),
            peer_links=frozenset(peer_links),
            ixp_asns=frozenset(ixp_asns),
            input_clique_asns=frozenset(input_clique_asns),
        )

    def _get_as_graph(self, as_graph_info: ASGraphInfo) -> ASGraph:
        """Creates and returns the ASGraph"""

        return self.ASGraphCls(as_graph_info, **self.as_graph_kwargs)

    #################
    # Parsing funcs #
    #################

    def _extract_input_clique_asns(
        self, line: str, input_clique_asns: set[int], invalid_asns: frozenset[int]
    ) -> None:
        """Adds all ASNs within input clique line to ases dict"""

        # Gets all input ASes for clique
        for asn in line.split(":")[-1].strip().split(" "):
            # Insert AS into graph
            if int(asn) not in invalid_asns:
                input_clique_asns.add(int(asn))

    def _extract_ixp_asns(
        self, line: str, ixp_asns: set[int], invalid_asns: frozenset[int]
    ) -> None:
        """Adds all ASNs that are detected IXPs to ASes dict"""

        # Get all IXPs that Caida lists
        for asn in line.split(":")[-1].strip().split(" "):
            if int(asn) not in invalid_asns:
                ixp_asns.add(int(asn))

    def _extract_provider_customers(
        self, line: str, cp_links: set[CPLink], invalid_asns: frozenset[int]
    ) -> None:
        """Extracts provider customers: <provider-as>|<customer-as>|-1"""

        provider_asn, customer_asn, _, source = line.split("|")
        if all(int(x) not in invalid_asns for x in (provider_asn, customer_asn)):
            cp_links.add(
                CPLink(customer_asn=int(customer_asn), provider_asn=int(provider_asn))
            )

    def _extract_peers(
        self, line: str, peer_links: set[PeerLink], invalid_asns: frozenset[int]
    ) -> None:
        """Extracts peers: <peer-as>|<peer-as>|0|<source>"""

        peer1_asn, peer2_asn, _, source = line.split("|")
        if all(int(x) not in invalid_asns for x in (peer1_asn, peer2_asn)):
            peer_links.add(PeerLink(int(peer1_asn), int(peer2_asn)))



from typing import TYPE_CHECKING
from warnings import warn

from bgpy.simulation_engine.ann_containers import RIBsIn, RIBsOut, SendQueue
from bgpy.simulation_engine.policies.bgp import BGP

from .process_incoming_funcs import (
    _new_ann_better,
    _process_incoming_withdrawal,
    _select_best_ribs_in,
    _withdraw_ann_from_neighbors,
    process_incoming_anns,
)
from .propagate_funcs import _prev_sent, _process_outgoing_ann, _propagate, _send_anns

if TYPE_CHECKING:
    from bgpy.shared.enums import Relationships
    from bgpy.simulation_engine.announcement import Announcement as Ann


class BGPFull(BGP):
    name = "BGP Full"

    def __init__(
        self,
        *args,
        ribs_in: RIBsIn | None = None,
        ribs_out: RIBsOut | None = None,
        send_q: SendQueue | None = None,
        **kwargs,
    ):
        super(BGPFull, self).__init__(*args, **kwargs)
        self.ribs_in: RIBsIn = ribs_in if ribs_in else RIBsIn()
        self.ribs_out: RIBsOut = ribs_out if ribs_out else RIBsOut()
        self.send_q: SendQueue = send_q if send_q else SendQueue()

    @property
    def _ribs_in(self) -> RIBsIn:
        warn(
            "Please use .ribs_in instead of ._ribs_in. "
            "This will be removed in a later version",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.ribs_in

    @property
    def _ribs_out(self) -> RIBsOut:
        warn(
            "Please use .ribs_out instead of ._ribs_out. "
            "This will be removed in a later version",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.ribs_out

    @property
    def _send_q(self) -> SendQueue:
        warn(
            "Please use .send_q instead of ._send_q. "
            "This will be removed in a later version",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.send_q

    # Propagation functions
    _propagate = _propagate
    _process_outgoing_ann = _process_outgoing_ann
    _prev_sent = _prev_sent
    _send_anns = _send_anns

    # Must add this func here since it refers to BGPFull
    # Could use super but want to avoid additional func calls
    def _populate_send_q(
        self,
        propagate_to: "Relationships",
        send_rels: set["Relationships"],
    ) -> None:
        # Process outging ann is oerriden so this just adds to send q
        super(BGPFull, self)._propagate(propagate_to, send_rels)

    # Process incoming funcs
    if not TYPE_CHECKING:
        process_incoming_anns = process_incoming_anns
        _new_ann_better = _new_ann_better
        _process_incoming_withdrawal = _process_incoming_withdrawal
        _withdraw_ann_from_neighbors = _withdraw_ann_from_neighbors
        _select_best_ribs_in = _select_best_ribs_in

    # NOTE: not sure why this is coded in such a weird fashion...
    def receive_ann(self, ann: "Ann", accept_withdrawals: bool = True) -> None:
        BGP.receive_ann(self, ann, accept_withdrawals=True)

    def __to_yaml_dict__(self):
        """This optional method is called when you call yaml.dump()"""

        as_dict = super(BGPFull, self).__to_yaml_dict__()
        as_dict.update(
            {
                "ribs_in": self.ribs_in,
                "ribs_out": self.ribs_out,
                "send_q": self.send_q,
            }
        )
        return as_dict



from frozendict import frozendict

from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import BGPFull, PeerROVFull
from bgpy.simulation_framework import ScenarioConfig, SubprefixHijack
from bgpy.tests.engine_tests.utils import EngineTestConfig

from .as_graph_info_000 import as_graph_info_000

desc = "Subprefix hijack with Peer ROV"

ex_config_007 = EngineTestConfig(
    name="ex_007_subprefix_hijack_peer_rov",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=SubprefixHijack,
        BasePolicyCls=BGPFull,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict({9: PeerROVFull}),
    ),
    as_graph_info=as_graph_info_000,
)



from .local_rib import LocalRIB
from .recv_queue import RecvQueue
from .ribs_in import AnnInfo, RIBsIn
from .ribs_out import RIBsOut
from .send_queue import SendInfo, SendQueue

__all__ = [
    "RecvQueue",
    "SendQueue",
    "SendInfo",
    "LocalRIB",
    "RIBsOut",
    "RIBsIn",
    "AnnInfo",
]



import math
import random
from dataclasses import replace
from functools import cached_property
from ipaddress import IPv4Network, IPv6Network, ip_network
from typing import TYPE_CHECKING, Any, Optional
from warnings import warn

from roa_checker import ROA

from bgpy.shared.enums import SpecialPercentAdoptions
from bgpy.simulation_engine import Announcement as Ann
from bgpy.simulation_engine import BaseSimulationEngine, Policy
from bgpy.simulation_framework.scenarios.scenario_config import ScenarioConfig

from .roa_helper_funcs import (
    _add_roa_info_to_anns,
    _get_roa_checker,
    _get_roa_origin,
    _get_roa_valid_length,
)

if TYPE_CHECKING:
    from bgpy.as_graphs import AS


class Scenario:
    """Contains information regarding a scenario/attack

    This represents a single trial and a single engine run

    NOTE: I wanted to split this class out into multiple files,
    but that will totally break mypy for subclassing. So for now,
    only the ROA funcs are moved out until this is fixed (since they
    are very unlikely to be subclassed).
    """

    min_propagation_rounds: int = 1

    def __init__(
        self,
        *,
        scenario_config: ScenarioConfig,
        percent_adoption: float | SpecialPercentAdoptions = 0,
        engine: BaseSimulationEngine | None = None,
        attacker_asns: frozenset[int] | None = None,
        victim_asns: frozenset[int] | None = None,
        adopting_asns: frozenset[int] | None = None,
    ):
        """inits attrs

        Any kwarg prefixed with default is only required for the test suite/YAML
        """

        # Config's ScenarioCls must be the same as instantiated Scenario
        assert scenario_config.ScenarioCls == self.__class__, (
            "The config's scenario class is "
            f"{scenario_config.ScenarioCls.__name__}, but the scenario used is "
            f"{self.__class__.__name__}"
        )

        self.scenario_config: ScenarioConfig = scenario_config
        self.percent_adoption: float | SpecialPercentAdoptions = percent_adoption

        self.attacker_asns: frozenset[int] = self._get_attacker_asns(
            scenario_config.override_attacker_asns,
            attacker_asns,
            engine,
        )

        self.victim_asns: frozenset[int] = self._get_victim_asns(
            scenario_config.override_victim_asns, victim_asns, engine
        )
        self.adopting_asns: frozenset[int] = self._get_adopting_asns(
            scenario_config.override_adopting_asns,
            adopting_asns,
            engine,
        )

        if self.scenario_config.override_announcements:
            anns = self.scenario_config.override_announcements
            self.roas: tuple[ROA, ...] = self.scenario_config.override_roas
            anns = self._add_roa_info_to_anns(announcements=anns, engine=engine)
            self.announcements: tuple[Ann, ...] = anns
        else:
            anns = self._get_announcements(engine=engine)
            self.roas = self._get_roas(announcements=anns, engine=engine)
            anns = self._add_roa_info_to_anns(announcements=anns, engine=engine)
            self.announcements = anns

        self.ordered_prefix_subprefix_dict: dict[str, list[str]] = (
            self._get_ordered_prefix_subprefix_dict()
        )

    #################
    # Get attackers #
    #################

    def _get_attacker_asns(
        self,
        override_attacker_asns: frozenset[int] | None,
        prev_attacker_asns: frozenset[int] | None,
        engine: BaseSimulationEngine | None,
    ) -> frozenset[int]:
        """Returns attacker ASN at random"""

        # This is coming from YAML, do not recalculate
        if override_attacker_asns is not None:
            attacker_asns = override_attacker_asns
            branch = 0
        # Reuse the attacker from the last scenario for comparability
        elif (
            prev_attacker_asns
            and len(prev_attacker_asns) == self.scenario_config.num_attackers
        ):
            attacker_asns = prev_attacker_asns
            branch = 1
        # If old scenario has less attackers, get a superset of their attackers
        elif (
            prev_attacker_asns
            and len(prev_attacker_asns) < self.scenario_config.num_attackers
        ):
            branch = 2
            assert engine
            possible_attacker_asns = self._get_possible_attacker_asns(
                engine, self.percent_adoption
            )
            possible_attacker_asns = possible_attacker_asns.difference(
                prev_attacker_asns
            )

            assert len(possible_attacker_asns) >= (
                self.scenario_config.num_attackers - len(prev_attacker_asns)
            )
            # https://stackoverflow.com/a/15837796/8903959
            new_attacker_asns = frozenset(
                random.sample(
                    tuple(possible_attacker_asns),
                    self.scenario_config.num_attackers - len(prev_attacker_asns),
                )
            )
            attacker_asns = prev_attacker_asns | new_attacker_asns

        # This is being initialized for the first time
        else:
            branch = 3
            assert engine
            possible_attacker_asns = self._get_possible_attacker_asns(
                engine, self.percent_adoption
            )

            assert len(possible_attacker_asns) >= self.scenario_config.num_attackers
            # https://stackoverflow.com/a/15837796/8903959
            attacker_asns = frozenset(
                random.sample(
                    tuple(possible_attacker_asns), self.scenario_config.num_attackers
                )
            )

        # Validate attacker asns
        err = f"Number of attackers is different from attacker length: Branch {branch}"
        assert len(attacker_asns) == self.scenario_config.num_attackers, err

        return attacker_asns

    def _get_possible_attacker_asns(
        self,
        engine: BaseSimulationEngine,
        percent_adoption: float | SpecialPercentAdoptions,
    ) -> frozenset[int]:
        """Returns possible attacker ASNs, defaulted from config"""

        possible_asns = engine.as_graph.asn_groups[
            self.scenario_config.attacker_subcategory_attr
        ]
        err = "Make mypy happy"
        assert all(isinstance(x, int) for x in possible_asns), err
        assert isinstance(possible_asns, frozenset), err
        return possible_asns

    ###############
    # Get Victims #
    ###############

    def _get_victim_asns(
        self,
        override_victim_asns: frozenset[int] | None,
        prev_victim_asns: frozenset[int] | None,
        engine: BaseSimulationEngine | None,
    ) -> frozenset[int]:
        """Returns victim ASN at random"""

        # This is coming from YAML, do not recalculate
        if override_victim_asns is not None:
            victim_asns = override_victim_asns
        # Reuse the victim from the last scenario for comparability
        elif (
            prev_victim_asns
            and len(prev_victim_asns) == self.scenario_config.num_victims
        ):
            victim_asns = prev_victim_asns
        # This is being initialized for the first time
        else:
            assert engine
            possible_victim_asns = self._get_possible_victim_asns(
                engine, self.percent_adoption
            )
            # https://stackoverflow.com/a/15837796/8903959
            victim_asns = frozenset(
                random.sample(
                    tuple(possible_victim_asns), self.scenario_config.num_victims
                )
            )

        err = "Number of victims is different from victim length"
        assert len(victim_asns) == self.scenario_config.num_victims, err

        return victim_asns

    def _get_possible_victim_asns(
        self,
        engine: BaseSimulationEngine,
        percent_adoption: float | SpecialPercentAdoptions,
    ) -> frozenset[int]:
        """Returns possible victim ASNs, defaulted from config"""

        possible_asns = engine.as_graph.asn_groups[
            self.scenario_config.victim_subcategory_attr
        ]
        err = "Make mypy happy"
        assert all(isinstance(x, int) for x in possible_asns), err
        assert isinstance(possible_asns, frozenset), err
        # Remove attackers from possible victims
        possible_asns = possible_asns.difference(self.attacker_asns)
        return possible_asns

    #######################
    # Adopting ASNs funcs #
    #######################

    def _get_adopting_asns(
        self,
        override_adopting_asns: frozenset[int] | None,
        adopting_asns: frozenset[int] | None,
        engine: BaseSimulationEngine | None,
    ) -> frozenset[int]:
        """Returns all asns that will be adopting self.AdoptPolicyCls"""

        if override_adopting_asns is not None:
            return override_adopting_asns
        # By default use the same adopting ASes as the last scenario config
        elif adopting_asns:
            return adopting_asns
        else:
            assert engine, "either yaml or engine must be set"
            adopting_asns = self._get_randomized_adopting_asns(engine)

        return adopting_asns

    def _get_randomized_adopting_asns(
        self,
        engine: BaseSimulationEngine,
    ) -> frozenset[int]:
        """Returns the set of adopting ASNs

        NOTE: this doesn't include ASNs listed in:
        hardcoded_asn_cls_dict
        default_adopters
        """

        adopting_asns: list[int] = list()
        # Randomly adopt in all three subcategories
        for subcategory in self.scenario_config.adoption_subcategory_attrs:
            asns = engine.as_graph.asn_groups[subcategory]
            # Remove ASes that are already pre-set
            # Ex: Attackers and victims, self.scenario_config.hardcoded_asn_cls_dict
            possible_adopters = asns.difference(self._preset_asns)

            # Get how many ASes should be adopting (store as k)
            if self.percent_adoption == SpecialPercentAdoptions.ONLY_ONE:
                k = 1
            elif self.percent_adoption == SpecialPercentAdoptions.ALL_BUT_ONE:
                k = len(possible_adopters) - 1
            # Really used just for testing
            elif self.percent_adoption == 0:
                k = 0
            else:
                err = f"{self.percent_adoption}"
                assert isinstance(self.percent_adoption, float), err
                k = math.ceil(len(possible_adopters) * self.percent_adoption)

            try:
                # https://stackoverflow.com/a/15837796/8903959
                adopting_asns.extend(random.sample(tuple(possible_adopters), k))
            except ValueError as e:
                raise ValueError(
                    f"{k} can't be sampled from {len(possible_adopters)}"
                ) from e
        return frozenset(adopting_asns)

    @cached_property
    def _default_adopters(self) -> frozenset[int]:
        """By default, victim always adopts"""

        return self.victim_asns

    @cached_property
    def _default_non_adopters(self) -> frozenset[int]:
        """By default, attacker always does not adopt"""

        return self.attacker_asns

    @cached_property
    def _preset_asns(self) -> frozenset[int]:
        """ASNs that have a preset adoption policy"""

        # Returns the union of default adopters and non adopters
        hardcoded_asns = set(self.scenario_config.hardcoded_asn_cls_dict)
        return self._default_adopters | self._default_non_adopters | hardcoded_asns

    @property
    def untracked_asns(self) -> frozenset[int]:
        """Returns ASNs that shouldn't be tracked by the metric tracker

        By default just the default adopters and non adopters
        """

        return self._default_adopters | self._default_non_adopters

    @property
    def _untracked_asns(self) -> frozenset[int]:
        """Returns ASNs that shouldn't be tracked by the metric tracker

        By default just the default adopters and non adopters
        """

        warn(
            "Instead of ._untracked_asns, please use .untracked_asns",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.untracked_asns

    #############################
    # Engine Manipulation Funcs #
    #############################

    def setup_engine(self, engine: BaseSimulationEngine) -> None:
        """Sets up engine

        Left this here so that it can be used as a hook
        """

        engine.setup(self)

    def get_policy_cls(self, as_obj: "AS") -> type[Policy]:
        """Returns the policy class for a given AS to set"""

        asn = as_obj.asn
        if self.scenario_config.AttackerBasePolicyCls and asn in self.attacker_asns:
            return self.scenario_config.AttackerBasePolicyCls
        elif asn in self._default_adopters:
            return self.scenario_config.AdoptPolicyCls
        elif Cls := self.scenario_config.hardcoded_asn_cls_dict.get(asn):
            return Cls
        elif asn in self.adopting_asns:
            return self.scenario_config.AdoptPolicyCls
        else:
            return self.scenario_config.BasePolicyCls

    ##################
    # Subclass Funcs #
    ##################

    def _get_announcements(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple["Ann", ...]:
        """Returns announcements

        Empty by default for testing, typically subclassed
        """

        return ()

    def _get_roas(
        self,
        *,
        announcements: tuple["Ann", ...] = (),
        engine: BaseSimulationEngine | None = None,
    ) -> tuple[ROA, ...]:
        """Returns a tuple of ROA's

        Not abstract and by default does nothing for
        backwards compatability
        """
        return ()

    def pre_aggregation_hook(
        self,
        engine: "BaseSimulationEngine",
        percent_adopt: float | SpecialPercentAdoptions,
        trial: int,
        propagation_round: int,
    ) -> None:
        """Useful hook for changes/checks
        prior to results aggregation.
        """
        pass

    def post_propagation_hook(
        self,
        engine: "BaseSimulationEngine",
        percent_adopt: float | SpecialPercentAdoptions,
        trial: int,
        propagation_round: int,
    ) -> None:
        """Useful hook for post propagation"""

        pass

    ####################
    # ROA Helper funcs #
    ####################

    _add_roa_info_to_anns = _add_roa_info_to_anns
    _get_roa_checker = _get_roa_checker
    _get_roa_origin = _get_roa_origin
    _get_roa_valid_length = _get_roa_valid_length

    ################
    # Helper Funcs #
    ################

    def _get_ordered_prefix_subprefix_dict(self):
        """Saves a dict of prefix to subprefixes"""

        prefixes = set()
        for ann in self.announcements:
            prefixes.add(ann.prefix)
        # Add ROA prefixes here, so that if we blackhole a non routed
        # prefix of a superprefix hijack this won't break
        # (since the prefix would only exist in the ROA)
        for roa in self.roas:
            prefixes.add(str(roa.prefix))
        # Do this here for speed
        prefixes: list[IPv4Network | IPv6Network] = [ip_network(x) for x in prefixes]
        # Sort prefixes with most specific prefix first
        # Note that this must be sorted for the traceback to get the
        # most specific prefix first
        prefixes = sorted(prefixes, key=lambda x: x.num_addresses)

        prefix_subprefix_dict = {x: [] for x in prefixes}
        for outer_prefix, subprefix_list in prefix_subprefix_dict.items():
            for prefix in prefixes:
                if prefix.subnet_of(outer_prefix) and prefix != outer_prefix:
                    subprefix_list.append(str(prefix))
        # Get rid of ip_network
        return {str(k): v for k, v in prefix_subprefix_dict.items()}

    ##############
    # Yaml Funcs #
    ##############

    def __to_yaml_dict__(self) -> dict[Any, Any]:
        """This optional method is called when you call yaml.dump()"""

        config_to_save = replace(
            self.scenario_config,
            override_attacker_asns=self.attacker_asns,
            override_victim_asns=self.victim_asns,
            override_adopting_asns=self.adopting_asns,
            override_announcements=self.announcements,
        )

        return {
            "scenario_config": config_to_save,
            "percent_adoption": self.percent_adoption,
        }

    @classmethod
    def __from_yaml_dict__(cls, dct, yaml_tag):
        """This optional method is called when you call yaml.load()"""

        return cls(
            scenario_config=dct["scenario_config"],
            percent_adoption=dct["percent_adoption"],
        )



from .bgpisec_transitive import BGPiSecTransitive
from .bgpisec_transitive_only_to_customers import BGPiSecTransitiveOnlyToCustomers
from .bgpisec_transitive_pro_con_id import BGPiSecTransitiveProConID
from .bgpisec import BGPiSec
from .provider_cone_id import ProviderConeID
from .bgpisec_transitive_full import BGPiSecTransitiveFull
from .bgpisec_transitive_only_to_customers_full import (
    BGPiSecTransitiveOnlyToCustomersFull,
)
from .bgpisec_transitive_pro_con_id_full import BGPiSecTransitiveProConIDFull
from .bgpisec_full import BGPiSecFull
from .provider_cone_id_full import ProviderConeIDFull


__all__ = [
    "BGPiSecTransitive",
    "BGPiSecTransitiveOnlyToCustomers",
    "BGPiSecTransitiveProConID",
    "BGPiSec",
    "ProviderConeID",
    "BGPiSecTransitiveFull",
    "BGPiSecTransitiveOnlyToCustomersFull",
    "BGPiSecTransitiveProConIDFull",
    "BGPiSecFull",
    "ProviderConeIDFull",
]



import subprocess
import sys
from pathlib import Path

import pytest

from bgpy.as_graphs import CAIDAASGraphCollector
from bgpy.shared.constants import bgpy_logger
from bgpy.tests.engine_tests import DiagramAggregator

DIAGRAM_PATH = Path(__file__).parent / "engine_tests" / "engine_test_outputs"


# https://github.com/pytest-dev/pytest-xdist/issues/783
def pytest_configure(config):
    """Caches the caida collector before parallelization"""

    # Prevent workers from running the same code
    if not hasattr(config, "workerinput"):
        # Caches CAIDA downloaded file only once before tests run
        CAIDAASGraphCollector().run()


def pytest_sessionfinish(session, exitstatus):
    """Runs when all tests are done, once per session, even with xdist

    https://github.com/pytest-dev/pytest-xdist/issues/271#issuecomment-826396320  # noqa
    """

    # Only run in master thread after all other threads/tests have finished
    # Also runs when xdist isn't running
    if not hasattr(session.config, "workerinput"):
        try:
            DiagramAggregator(DIAGRAM_PATH).aggregate_diagrams()
            # Teardown stuff (open PDF for viewing)
            if session.config.getoption("view"):
                # https://stackoverflow.com/q/19453338/8903959
                agg_path = DiagramAggregator(DIAGRAM_PATH).aggregated_diagrams_path
                command = "open" if sys.platform == "darwin" else "xdg-open"
                subprocess.call([command, str(agg_path)])  # noqa: S603
        # If the diagram aggregator errors out due to a test failure,
        # the actual test failure is suppressed for some reason and only the diagram
        # aggregator error is raised (likely due to some pytest internals)
        # I've never encountered a legitimate error with the PDF generation, and if
        # the PDF generation fails, but the tests pass, I think it's okay to pass
        # If there's a better workaround for this, I don't know it.
        except Exception as e:  # noqa: BLE001
            bgpy_logger.exception(
                f"Diagram aggregator failed, typically caused by a test failure {e}"
            )


@pytest.fixture(scope="session")
def overwrite(pytestconfig):
    return pytestconfig.getoption("overwrite")


@pytest.fixture(scope="session")
def dpi(pytestconfig):
    return pytestconfig.getoption("dpi")


# https://stackoverflow.com/a/66597438/8903959
def pytest_addoption(parser):
    # View test PDF when complete
    parser.addoption(
        "--view",
        action="store_true",
        default=False,
        help="View the diagrams in a PDF after tests run",
    )
    # Overwrite ground truth
    parser.addoption(
        "--overwrite",
        action="store_true",
        default=False,
        help="Overwrite the ground truths of the tests",
    )
    # Store DPI
    parser.addoption(
        "--dpi",
        "--DPI",
        action="store",
        type=int,
        default=96,
        help="Store the DPI to render the diagrams",
    )



from bgpy.as_graphs import ASGraphInfo
from bgpy.as_graphs.base.links import CustomerProviderLink as CPLink
from bgpy.as_graphs.base.links import PeerLink
from bgpy.shared.enums import ASNs

r"""

This graph can highlight every example of Gao Rexford Valley Free Routing

(see config 001 for description)
Graph is too complex for an ascii drawing, just view the PDF
"""


as_graph_info_000 = ASGraphInfo(
    peer_links=frozenset(
        {
            PeerLink(8, 9),
            PeerLink(9, 10),
            PeerLink(9, 3),
        }
    ),
    customer_provider_links=frozenset(
        [
            CPLink(provider_asn=1, customer_asn=ASNs.ATTACKER.value),
            CPLink(provider_asn=2, customer_asn=ASNs.ATTACKER.value),
            CPLink(provider_asn=2, customer_asn=ASNs.VICTIM.value),
            CPLink(provider_asn=4, customer_asn=ASNs.VICTIM.value),
            CPLink(provider_asn=5, customer_asn=1),
            # CPLink(provider_asn=5, customer_asn=2),
            CPLink(provider_asn=8, customer_asn=1),
            CPLink(provider_asn=8, customer_asn=2),
            CPLink(provider_asn=9, customer_asn=4),
            CPLink(provider_asn=10, customer_asn=ASNs.VICTIM.value),
            CPLink(provider_asn=11, customer_asn=8),
            CPLink(provider_asn=11, customer_asn=9),
            CPLink(provider_asn=11, customer_asn=10),
            CPLink(provider_asn=12, customer_asn=10),
        ]
    ),
    diagram_ranks=(
        (ASNs.ATTACKER.value, ASNs.VICTIM.value),
        (1, 2, 3, 4),
        (5, 8, 9, 10),
        (11, 12),
    ),
)



import gc
from statistics import mean

import matplotlib.pyplot as plt
from matplotlib.container import ErrorbarContainer


def _add_legends_and_save(
    self,
    fig,
    ax,
    graph_category,
    non_aggregated_data_dict,
    max_attacker_dict,
    graph_name,
):
    first_legend, aggregated_labels_handles_dict = self._add_legend(
        fig,
        ax,
        graph_category,
        non_aggregated_data_dict,
        max_attacker_dict,
    )

    self._add_strongest_attacker_legend(
        fig,
        ax,
        graph_category,
        non_aggregated_data_dict,
        max_attacker_dict,
        first_legend,
        aggregated_labels_handles_dict,
    )

    self._save_and_close_graph(fig, ax, graph_name)


def _add_legend(
    self,
    fig,
    ax,
    graph_category,
    non_aggregated_data_dict,
    max_attacker_data_dict,
):
    """Add legend to the graph"""

    # This is to avoid warnings
    handles, labels = ax.get_legend_handles_labels()

    # non aggregated
    non_aggregated_labels = {x.label for x in non_aggregated_data_dict.values()}
    non_aggregated_labels_handles_dict = dict()
    aggregated_labels_handles_dict = dict()
    for handle, label in zip(handles, labels, strict=False):
        # This is a placeholder line, don't plot in the legend
        if label in non_aggregated_labels:
            non_aggregated_labels_handles_dict[label] = handle
        else:
            aggregated_labels_handles_dict[label] = handle

    mean_y_dict = {
        line_data.line_info.label: mean(line_data.ys)
        for label, line_data in non_aggregated_data_dict.items()
    }
    sorted_labels = sorted(
        non_aggregated_labels_handles_dict,
        key=lambda label: mean_y_dict[label],
        reverse=True,
    )
    sorted_handles = [non_aggregated_labels_handles_dict[lbl] for lbl in sorted_labels]

    first_legend = ax.legend(sorted_handles, sorted_labels)
    return first_legend, aggregated_labels_handles_dict


def _add_strongest_attacker_legend(
    self,
    fig,
    ax,
    graph_category,
    non_aggregated_data_dict,
    max_attacker_data_dict,
    first_legend,
    aggregated_labels_handles_dict,
):
    """Add second legend for strongest attacker"""

    # Only run if there is a secondary strongest attacker
    if not self.strongest_attacker_dict:
        return

    ax.add_artist(first_legend)
    # First you must draw the legend so that you can get it's location
    plt.tight_layout()
    # https://stackoverflow.com/a/43748841/8903959
    fig.canvas.draw()

    bbox = first_legend.get_window_extent()
    # Transform the bounding box into figure coordinates
    bbox_transformed = bbox.transformed(fig.transFigure.inverted())

    label_to_legend_label_dict = dict()
    for line_info_tup in self.strongest_attacker_dict.values():
        for line_info in line_info_tup:
            label_to_legend_label_dict[line_info.label] = (
                line_info.strongest_attacker_legend_label
            )

    # This way the same handle isn't repeated multiple times
    handles = list()
    used_labels = list()
    for label, handle in aggregated_labels_handles_dict.items():
        label_to_use = label_to_legend_label_dict[label]
        if label_to_use not in used_labels:
            used_labels.append(label_to_use)

            # Modify the handle to make its markers grey
            if isinstance(handle, ErrorbarContainer):
                # Create a custom line with grey marker for the legend
                handles.append(
                    plt.Line2D(  # type: ignore
                        [],
                        [],
                        color=handle.lines[0].get_color(),
                        marker=handle.lines[0].get_marker(),
                        linestyle="",
                        markerfacecolor="grey",
                        markeredgecolor="grey",
                    )
                )
            else:
                raise NotImplementedError

    ax.legend(
        labels=used_labels,
        handles=handles,
        title="Strongest Attacker",
        # Upper right corner should line up
        loc="upper right",
        bbox_to_anchor=(bbox_transformed.x1, bbox_transformed.y0),
        bbox_transform=fig.transFigure,
    )


def _save_and_close_graph(self, fig, ax, graph_name):
    """Saves and closes the graph"""

    path = self.graph_dir / graph_name
    path.parent.mkdir(parents=True, exist_ok=True)
    plt.savefig(path)
    # https://stackoverflow.com/a/33343289/8903959
    ax.cla()
    plt.cla()
    plt.clf()
    # If you just close the fig, on machines with many CPUs and trials,
    # there is some sort of a memory leak that occurs. See stackoverflow
    # comment above
    plt.close(fig)
    # If you are running one simulation after the other, matplotlib
    # basically leaks memory. I couldn't find the original issue, but
    # here is a note in one of their releases saying to just call the garbage
    # collector: https://matplotlib.org/stable/users/prev_whats_new/
    # whats_new_3.6.0.html#garbage-collection-is-no-longer-run-on-figure-close
    # and here is the stackoverflow post on this topic:
    # https://stackoverflow.com/a/33343289/8903959
    # Even if this works without garbage collection in 3.5.2, that will break
    # as soon as we upgrade to the latest matplotlib which no longer does
    # If you run the simulations on a machine with many cores and lots of trials,
    # this bug leaks enough memory to crash the server, so we must garbage collect
    gc.collect()



from typing import TYPE_CHECKING, Optional

from bgpy.shared.enums import Prefixes, Timestamps
from bgpy.simulation_framework.scenarios.custom_scenarios.victims_prefix import (
    VictimsPrefix,
)

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann
    from bgpy.simulation_engine import BaseSimulationEngine


class SubprefixHijack(VictimsPrefix):
    """Subprefix Hijack Scenario

    Subprefix hijack consists of a valid prefix by the victim with a roa
    then a subprefix from an attacker
    invalid by roa by length and origin
    """

    def _get_announcements(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple["Ann", ...]:
        """Returns victim and attacker anns for subprefix hijack

        for subclasses of this EngineInput, you can set AnnCls equal to
        something other than Announcement
        """

        # First get victim's anns
        victim_anns = super()._get_announcements(engine=engine)
        assert isinstance(victim_anns, tuple), "mypy"
        attacker_anns = self._get_subprefix_attacker_anns(engine=engine)
        return victim_anns + attacker_anns

    def _get_subprefix_attacker_anns(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple["Ann", ...]:
        """Returns subprefix announcements from the attacker"""

        anns = list()
        for attacker_asn in self.attacker_asns:
            anns.append(
                self.scenario_config.AnnCls(
                    prefix=Prefixes.SUBPREFIX.value,
                    as_path=(attacker_asn,),
                    timestamp=Timestamps.ATTACKER.value,
                )
            )
        return tuple(anns)



from frozendict import frozendict

from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import ROV
from bgpy.simulation_framework import FirstASNStrippingPrefixHijack, ScenarioConfig
from bgpy.tests.engine_tests.utils import EngineTestConfig

from .as_graph_info_000 import as_graph_info_000

desc = (
    "Nieghbor spoofing prefix hijack thwarting ROV\n"
    "This attack reaches more ASes than just the origin hijack"
)

ex_config_013 = EngineTestConfig(
    name="ex_013_neighbor_spoofing_hijack_rov_simple",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=FirstASNStrippingPrefixHijack,
        BasePolicyCls=ROV,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict({}),
    ),
    as_graph_info=as_graph_info_000,
)



from typing import TYPE_CHECKING, Optional

from bgpy.shared.enums import Prefixes, Timestamps
from bgpy.simulation_framework.scenarios.custom_scenarios.victims_prefix import (
    VictimsPrefix,
)

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann
    from bgpy.simulation_engine import BaseSimulationEngine


class ForgedOriginPrefixHijack(VictimsPrefix):
    """Extension of prefix hijack where attacker appends legit origin to avoid ROV

    Same ROAs as the VictimsPrefix, which is why we subclassed it it
    """

    def _get_announcements(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple["Ann", ...]:
        """Returns the two announcements seeded for this engine input

        This engine input is for a prefix hijack,
        consisting of a valid prefix and invalid prefix with path manipulation
        """

        # First get the victims prefix
        victim_anns = super()._get_announcements(engine=engine)
        attacker_anns = self._get_forged_origin_attacker_anns(engine=engine)
        return victim_anns + attacker_anns

    def _get_forged_origin_attacker_anns(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple["Ann", ...]:
        """Returns attacker anns for a forged origin hijack"""

        victim_asn = next(iter(self.victim_asns))

        anns = list()
        for attacker_asn in self.attacker_asns:
            anns.append(
                self.scenario_config.AnnCls(
                    prefix=Prefixes.PREFIX.value,
                    as_path=(attacker_asn, victim_asn),
                    timestamp=Timestamps.ATTACKER.value,
                    next_hop_asn=attacker_asn,
                    seed_asn=attacker_asn,
                )
            )
        return tuple(anns)



from typing import TYPE_CHECKING, Optional

from bgpy.as_graphs import AS
from bgpy.shared.enums import Outcomes, Plane, Relationships
from bgpy.simulation_engine import Announcement as Ann
from bgpy.simulation_engine import BaseSimulationEngine

from .base_as_graph_analyzer import BaseASGraphAnalyzer

if TYPE_CHECKING:
    from bgpy.simulation_framework.scenarios import Scenario


class ASGraphAnalyzer(BaseASGraphAnalyzer):
    """Takes in a BaseSimulationEngine and outputs metrics"""

    def __init__(
        self,
        engine: BaseSimulationEngine,
        scenario: "Scenario",
        data_plane_tracking: bool = True,
        control_plane_tracking: bool = False,
        # Most to least specific
        ordered_prefixes: tuple[str, ...] = (),
    ) -> None:
        self.engine: BaseSimulationEngine = engine
        self.scenario: Scenario = scenario
        if not ordered_prefixes:
            ordered_prefixes = tuple(self.scenario.ordered_prefix_subprefix_dict.keys())
        self._most_specific_ann_dict: dict[AS, Ann | None] = (
            self._get_most_specific_ann_dict(engine, ordered_prefixes)
        )
        self._data_plane_outcomes: dict[int, int] = dict()
        self._control_plane_outcomes: dict[int, int] = dict()
        self.outcomes: dict[int, dict[int, int]] = {
            Plane.DATA.value: self._data_plane_outcomes,
            Plane.CTRL.value: self._control_plane_outcomes,
        }
        self.data_plane_tracking: bool = data_plane_tracking
        self.control_plane_tracking: bool = control_plane_tracking

    def _get_most_specific_ann_dict(
        self, engine: BaseSimulationEngine, ordered_prefixes: tuple[str, ...]
    ) -> dict[AS, Optional["Ann"]]:
        """Gets the most specific ann in a list of ordered prefixes

        ordered prefixes start with the most specific, and move to least specific
        """

        return {
            x: self._get_most_specific_ann(x, ordered_prefixes) for x in engine.as_graph
        }

    def _get_most_specific_ann(
        self, as_obj: AS, ordered_prefixes: tuple[str, ...]
    ) -> Optional["Ann"]:
        """Returns the most specific announcement that exists in a rib

        as_obj is the as
        ordered prefixes are prefixes ordered from most specific to least
        """

        for prefix in ordered_prefixes:
            most_specific_ann = as_obj.policy.local_rib.get(prefix)
            if most_specific_ann:
                assert isinstance(most_specific_ann, Ann), "for mypy"
                return most_specific_ann
        return None

    def analyze(self) -> dict[int, dict[int, int]]:
        """Takes in engine and outputs traceback for ctrl + data plane data"""

        for as_obj in self.engine.as_graph:
            if self.data_plane_tracking:
                # Gets AS outcome and stores it in the outcomes dict
                self._get_as_outcome_data_plane(as_obj)
            if self.control_plane_tracking:
                self._get_as_outcome_ctrl_plane(as_obj)
            self._get_other_as_outcome_hook(as_obj)
        return self.outcomes

    ####################
    # Data plane funcs #
    ####################

    def _get_as_outcome_data_plane(self, as_obj: AS) -> int:
        """Recursively returns the as outcome"""

        if as_obj.asn in self._data_plane_outcomes:
            return self._data_plane_outcomes[as_obj.asn]
        else:
            most_specific_ann = self._most_specific_ann_dict[as_obj]
            outcome_int = self._determine_as_outcome_data_plane(
                as_obj, most_specific_ann
            )
            # We haven't traced back all the way on the AS path
            if outcome_int == Outcomes.UNDETERMINED.value:
                assert most_specific_ann, "If not disconnected, ann must exist"
                # next as in the AS path to traceback to
                # Ignore type because only way for this to be here
                # Is if the most specific "Ann" was NOT None.
                next_as = self.engine.as_graph.as_dict[
                    # NOTE: Starting in v4, this is the next hop,
                    # not the next ASN in the AS PATH
                    # This is more in line with real BGP and allows for more
                    # advanced types of hijacks such as origin spoofing hijacks
                    most_specific_ann.next_hop_asn
                ]
                outcome_int = self._get_as_outcome_data_plane(next_as)
            assert outcome_int != Outcomes.UNDETERMINED.value, "Shouldn't be possible"

            self._data_plane_outcomes[as_obj.asn] = outcome_int
            return outcome_int

    def _determine_as_outcome_data_plane(
        self, as_obj: AS, most_specific_ann: Optional["Ann"]
    ) -> int:
        """Determines the outcome at an AS

        ann is most_specific_ann is the most specific prefix announcement
        that exists at that AS
        """

        if as_obj.asn in self.scenario.attacker_asns:
            return Outcomes.ATTACKER_SUCCESS.value
        elif as_obj.asn in self.scenario.victim_asns:
            return Outcomes.VICTIM_SUCCESS.value
        # End of traceback
        elif (
            most_specific_ann is None
            or len(most_specific_ann.as_path) == 1
            or most_specific_ann.recv_relationship.value == Relationships.ORIGIN.value
            # Adding this condition in V4 for proper next_hop behavior
            or most_specific_ann.next_hop_asn == as_obj.asn
        ):
            return Outcomes.DISCONNECTED.value
        else:
            return Outcomes.UNDETERMINED.value

    #######################
    # Control Plane Funcs #
    #######################

    def _get_as_outcome_ctrl_plane(self, as_obj: AS) -> int:
        """Stores and returns the AS outcome from the control plane"""

        most_specific_ann = self._most_specific_ann_dict[as_obj]
        outcome_int = self._determine_as_outcome_ctrl_plane(as_obj, most_specific_ann)
        assert outcome_int != Outcomes.UNDETERMINED.value, "Shouldn't be possible"
        self._control_plane_outcomes[as_obj.asn] = outcome_int
        return outcome_int

    def _determine_as_outcome_ctrl_plane(self, as_obj: AS, ann: Optional["Ann"]) -> int:
        """Determines the outcome at an AS on the control plane

        ann is most_specific_ann is the most specific prefix announcement
        that exists at that AS
        """

        if not ann:
            return Outcomes.DISCONNECTED.value
        elif ann.origin in self.scenario.attacker_asns:
            return Outcomes.ATTACKER_SUCCESS.value
        elif ann.origin in self.scenario.victim_asns:
            return Outcomes.VICTIM_SUCCESS.value
        else:
            return Outcomes.DISCONNECTED.value

    ################################
    # Hook funcs for other metrics #
    ################################

    def _get_other_as_outcome_hook(self, as_obj: AS) -> int:
        # Noop, this is just to satisfy mypy
        return Outcomes.ATTACKER_SUCCESS.value
